from typing import Any

from pydantic import BaseModel, Field, ValidationInfo, field_validator
from pydantic_settings import BaseSettings, SettingsConfigDict

DEFAULT_POWERBI_SCOPES = ["https://analysis.windows.net/powerbi/api/.default"]
DEFAULT_GRAPH_SCOPES = ["https://graph.microsoft.com/.default"]
DEFAULT_SCOPES = DEFAULT_POWERBI_SCOPES + DEFAULT_GRAPH_SCOPES

POWERBI_ENV_PREFIX = "CASTOR_POWERBI_"

CLIENT_APP_BASE = "https://login.microsoftonline.com"
REST_API_BASE_PATH = "https://api.powerbi.com/v1.0/myorg"
GRAPH_API_BASE_PATH = "https://graph.microsoft.com/v1.0"


class PowerbiCertificate(BaseModel):
    """
    Supports all dict credentials formats supported by PowerBI
    https://learn.microsoft.com/en-us/python/api/msal/msal.application.confidentialclientapplication
    """

    client_assertion: str | None = None
    passphrase: str | None = None
    private_key: str | None = None
    private_key_pfx_path: str | None = None
    public_certificate: str | None = None
    thumbprint: str | None = None


class PowerbiCredentials(BaseSettings):
    """
    Class to handle authentication to both the Power BI API and
    the Microsoft Graph API
    """

    model_config = SettingsConfigDict(
        env_prefix=POWERBI_ENV_PREFIX,
        extra="ignore",
        populate_by_name=True,
    )

    client_id: str
    tenant_id: str
    secret: str | None = None
    certificate: PowerbiCertificate | None = None
    api_base: str = REST_API_BASE_PATH
    graph_api_base: str = GRAPH_API_BASE_PATH
    login_url: str = CLIENT_APP_BASE
    scopes: list[str] = Field(default_factory=lambda: DEFAULT_SCOPES)

    @field_validator(
        "api_base",
        "graph_api_base",
        "login_url",
        "scopes",
        mode="before",
    )
    @classmethod
    def set_defaults(cls, v: Any | None, info: ValidationInfo) -> Any:
        """
        Defaults need to be set explicitly due to a side effect of the
        `env_prefix` config.
        """
        if v is not None:
            return v

        defaults: dict[str, Any] = {
            "api_base": REST_API_BASE_PATH,
            "graph_api_base": GRAPH_API_BASE_PATH,
            "login_url": CLIENT_APP_BASE,
            "scopes": DEFAULT_SCOPES,
        }
        assert info.field_name
        return defaults[info.field_name]

    @property
    def powerbi_scopes(self) -> list[str]:
        return [scope for scope in self.scopes if "powerbi" in scope.lower()]

    @property
    def graph_scopes(self) -> list[str]:
        return [scope for scope in self.scopes if "graph" in scope.lower()]



import logging

from ...utils import LocalStorage, from_env, write_summary
from ..abstract import (
    CATALOG_ASSETS,
    EXTERNAL_LINEAGE_ASSETS,
    QUERIES_ASSETS,
    VIEWS_ASSETS,
    AssetsMapping,
    SQLExtractionProcessor,
    WarehouseAsset,
    WarehouseAssetGroup,
    common_args,
    extractable_assets,
)
from .client import MSSQLClient
from .query import MSSQLQueryBuilder

logger = logging.getLogger(__name__)


MSSQL_ASSETS: AssetsMapping = {
    WarehouseAssetGroup.CATALOG: CATALOG_ASSETS,
    WarehouseAssetGroup.EXTERNAL_LINEAGE: EXTERNAL_LINEAGE_ASSETS,
    WarehouseAssetGroup.QUERY: QUERIES_ASSETS,
    WarehouseAssetGroup.ROLE: (WarehouseAsset.USER,),
    WarehouseAssetGroup.VIEW_DDL: VIEWS_ASSETS,
}


MSSQL_USER = "CASTOR_MSSQL_USER"
MSSQL_PASSWORD = "CASTOR_MSSQL_PASSWORD"  # noqa: S105
MSSQL_HOST = "CASTOR_MSSQL_HOST"
MSSQL_PORT = "CASTOR_MSSQL_PORT"
MSSQL_DEFAULT_DB = "CASTOR_MSSQL_DEFAULT_DB"


def _credentials(params: dict) -> dict:
    """extract mssql credentials"""
    return {
        "host": params.get("host") or from_env(MSSQL_HOST),
        "password": params.get("password") or from_env(MSSQL_PASSWORD),
        "port": params.get("port") or from_env(MSSQL_PORT),
        "user": params.get("user") or from_env(MSSQL_USER),
        "default_db": params.get("default_db")
        or from_env(MSSQL_DEFAULT_DB, allow_missing=True),
    }


def extract_all(**kwargs) -> None:
    """
    Extract all assets from mssql and store the results in CSV files
    """
    output_directory, skip_existing = common_args(kwargs)

    client = MSSQLClient(
        credentials=_credentials(kwargs),
        db_allowed=kwargs.get("db_allowed"),
        db_blocked=kwargs.get("db_blocked"),
    )

    databases = client.get_databases()
    query_databases = client.get_databases(with_query_store_enabled=True)

    query_builder = MSSQLQueryBuilder(
        databases=databases,
        query_databases=query_databases,
    )

    storage = LocalStorage(directory=output_directory)
    extractor = SQLExtractionProcessor(
        client=client,
        query_builder=query_builder,
        storage=storage,
    )

    skip_queries = kwargs.get("skip_queries") or False
    skip_groups = {WarehouseAssetGroup.QUERY} if skip_queries else None

    for asset in extractable_assets(MSSQL_ASSETS, skip_groups=skip_groups):
        logger.info(f"Extracting `{asset.value.upper()}` ...")
        location = extractor.extract(asset, skip_existing)
        logger.info(f"Results stored to {location}\n")

    write_summary(
        output_directory,
        storage.stored_at_ts,
        client_name=client.name(),
    )



from collections.abc import Iterable, Iterator
from functools import partial
from http import HTTPStatus

from requests import HTTPError

from ....utils import (
    APIClient,
    RequestSafeMode,
    fetch_all_pages,
)
from .authentication import PowerBiBearerAuth
from .credentials import PowerbiCredentials
from .endpoints import GraphAPIEndpointFactory
from .pagination import GraphAPIPagination

VOLUME_IGNORED = 100
# there are special users (e.g. Tenant Administrators) that do not exist
# outside of Power BI, and hence cannot be fetched from Graph API
IGNORED_ERROR_CODES = (HTTPStatus.NOT_FOUND,)
GRAPH_SAFE_MODE = RequestSafeMode(
    max_errors=VOLUME_IGNORED,
    status_codes=IGNORED_ERROR_CODES,
)


class MicrosoftGraphAccessForbidden(Exception):
    def __init__(self):
        error_msg = (
            "The user does not have permissions to access Microsoft Graph."
        )
        super().__init__(error_msg)


class MicrosoftGraphPIClient(APIClient):
    def __init__(
        self,
        credentials: PowerbiCredentials,
    ):
        auth = PowerBiBearerAuth(credentials=credentials, scope_type="graph")
        super().__init__(
            auth=auth,
            safe_mode=GRAPH_SAFE_MODE,
        )
        self.endpoint_factory = GraphAPIEndpointFactory(
            login_url=credentials.login_url,
            api_base=credentials.graph_api_base,
        )

    def _group_members(self, group_id: str) -> list[dict]:
        """Get all transitive members of a single group"""
        request = partial(
            self._get,
            endpoint=self.endpoint_factory.transitive_group_members(group_id),
        )
        users = list(fetch_all_pages(request, GraphAPIPagination))
        return users

    def users_in_groups(self, group_ids: Iterable[str]) -> Iterator[dict]:
        """
        Returns the list of users that are part of the given groups.
        This may contain duplicate entries.
        Raises a custom Exception if the user does not have permissions to
        read group members.
        """
        for group_id in group_ids:
            try:
                yield from self._group_members(group_id)
            except HTTPError as err:
                if err.response.status_code == HTTPStatus.FORBIDDEN:
                    raise MicrosoftGraphAccessForbidden()
                raise



import json

import pytest
from tableauserverclient.server.endpoint.exceptions import NonXMLResponseError

from .client_metadata_api import (
    _TIMEOUT_MESSAGES,
    _check_errors,
    _deduplicate,
    _format_filter_ids,
    _is_relationship_service_war_exception,
)
from .errors import TableauApiError, TableauApiTimeout


def test__deduplicate():
    result_pages = iter(
        [
            [
                {"id": 1, "name": "workbook_1"},
                {"id": 2, "name": "workbook_2"},
            ],
            [
                {"id": 1, "name": "workbook_1"},
                {"id": 3, "name": "workbook_3"},
                {"id": 4, "name": "workbook_4"},
            ],
            [
                {"id": 4, "name": "workbook_4"},
                {"id": 5, "name": "workbook_5"},
                {"id": 5, "name": "workbook_5"},
                {"id": 5, "name": "workbook_5"},
            ],
            [
                {"id": 1, "name": "workbook_1"},
                {"id": 3, "name": "workbook_3"},
            ],
        ]
    )
    deduplicated = _deduplicate(result_pages)
    assert len(deduplicated) == 5
    deduplicated_keys = {item["id"] for item in deduplicated}
    assert deduplicated_keys == {1, 2, 3, 4, 5}


def _timeout_error():
    return {"message": _TIMEOUT_MESSAGES[0]}


def _critical_error():
    return {
        "message": "critical error",
        "extensions": {"severity": "Error"},
    }


def _unknown_severity_error():
    return {
        "message": "unknown severity error",
        "extensions": {"severity": None},
    }


def _warning():
    return {
        "message": "critical error",
        "extensions": {"severity": "Warning"},
    }


def test__check_errors():
    answer = {"data": []}  # no errors
    assert _check_errors(answer) is None

    # only warnings
    answer = {
        "data": [],
        "errors": [
            _warning(),
            _warning(),
            _warning(),
        ],
    }
    assert _check_errors(answer) is None

    # timeout issues should be prio (because they can be retried)
    answer = {
        "data": [],
        "errors": [
            _warning(),
            _critical_error(),
            _timeout_error(),
            _warning(),
        ],
    }
    with pytest.raises(TableauApiTimeout):
        _check_errors(answer)

    # expect critical error
    answer = {
        "data": [],
        "errors": [
            _warning(),
            _warning(),
            _critical_error(),
        ],
    }
    with pytest.raises(TableauApiError):
        _check_errors(answer)

    # unknown severity is considered as critical
    answer = {
        "data": [],
        "errors": [
            _unknown_severity_error(),
            _warning(),
        ],
    }
    with pytest.raises(TableauApiError):
        _check_errors(answer)


def test__format_filter_ids():
    expected = '{ idWithin: ["foo"] }'
    assert _format_filter_ids(["foo"]) == expected

    expected = '{ idWithin: ["foo", "bar"] }'
    assert _format_filter_ids(["foo", "bar"]) == expected


def _non_xml_error(payload: dict):
    error_bytes = json.dumps(payload).encode("utf-8")
    exception = NonXMLResponseError(error_bytes)
    return exception


def test__is_relationship_service_war_exception():
    # happy path
    payload = {
        "status": 401,
        "path": "/relationship-service-war/graphql",
    }
    exception = _non_xml_error(payload)
    assert _is_relationship_service_war_exception(exception) is True

    # another type of NonXmlError
    payload = {
        "status": 500,
        "path": "internal_error/graphql",
    }
    exception = _non_xml_error(payload)
    assert _is_relationship_service_war_exception(exception) is False

    # empty payload
    payload = {}
    exception = _non_xml_error(payload)
    assert _is_relationship_service_war_exception(exception) is False

    # another type of exception
    exception = ValueError("foo")
    assert _is_relationship_service_war_exception(exception) is False



#!/usr/bin/env python3
import logging
import ntpath
from collections.abc import Iterable
from datetime import datetime
from uuid import UUID

import requests

from ..utils.retry import retry
from .constant import (
    INGEST_URLS,
    PATH_TEMPLATES,
    RETRY_BASE_MS,
    RETRY_JITTER_MS,
    RETRY_STRATEGY,
    FileType,
)
from .enums import Zone
from .env import get_blob_env
from .settings import UploaderSettings
from .utils import iter_files

logger = logging.getLogger(__name__)

_EXCEPTIONS = (
    requests.exceptions.Timeout,
    requests.exceptions.ConnectTimeout,
)


def _path_and_url(
    source_id: UUID,
    file_type: FileType,
    file_path: str,
    zone: Zone,
) -> tuple[str, str]:
    now = datetime.utcnow()
    timestamp = int(now.timestamp())
    filename = ntpath.basename(file_path)
    path_template = PATH_TEMPLATES[file_type]
    path = path_template.format(
        timestamp=timestamp,
        source_id=source_id,
        filename=filename,
    )

    url = f"{INGEST_URLS[zone]}/{path}"

    return path, url


def _headers(token: str) -> dict:
    return {
        "Authorization": f"Token {token}",
        "Accept": "text/csv, application/json",
    }


def _upload(
    token: str,
    source_id: UUID,
    file_path: str,
    file_type: FileType,
    zone: Zone | None = Zone.EU,
) -> None:
    """
    Upload the given file to Google Cloud Storage (GCS)
    - Don't call GCS API directly
    - Call the ingestion proxy which handles authorisation and uploading
    """
    if not zone:
        zone = Zone.EU
    path, url = _path_and_url(source_id, file_type, file_path, zone)
    headers = _headers(token)
    timeout, max_retries = get_blob_env()

    with open(file_path, "rb") as file_content:

        @retry(
            exceptions=_EXCEPTIONS,
            max_retries=max_retries,
            base_ms=RETRY_BASE_MS,
            jitter_ms=RETRY_JITTER_MS,
            strategy=RETRY_STRATEGY,
        )
        def _request_post():
            response = requests.post(
                url=url,
                headers=headers,
                files={"file": file_content},
                timeout=timeout,
            )
            response.raise_for_status()

        _request_post()

    logger.info(f"Uploaded {file_path} as {file_type.value} to {path}")


def upload_manifest(
    token: str,
    source_id: UUID,
    zone: Zone | None,
    file_path: str | None = None,
) -> None:
    """
    token: backend public API token
    source_id: id for the source
    file_path: path to the local manifest to upload
    """
    if not file_path:
        raise ValueError("file path is needed to upload a manifest")
    _upload(
        token=token,
        source_id=source_id,
        file_path=file_path,
        file_type=FileType.DBT,
        zone=zone,
    )


def upload(
    token: str,
    source_id: UUID,
    file_type: FileType,
    zone: Zone | None,
    file_path: str | None = None,
    directory_path: str | None = None,
) -> None:
    """
    token: backend public API token
    source_id: id for the source
    file_type: type of file(s) uploaded - see FileType Enum
    file_path: path to the local visualization or warehouse file to upload
    directory_path: path to the local directory containing files to upload
    """
    files: Iterable[str]
    if directory_path:
        files = iter_files(directory_path)
    elif file_path:
        files = [file_path]
    else:
        message = "either file_path or directory_path should be defined"
        raise ValueError(message)

    for file_ in files:
        _upload(
            token=token,
            source_id=source_id,
            file_path=file_,
            file_type=file_type,
            zone=zone,
        )


def upload_any(**kwargs) -> None:
    """
    entrypath to upload either a file or a manifest

    token: backend public API token
    source_id: id for the source
    file_type: type of file(s) uploaded - see FileType Enum
    file_path: path to the local visualization or warehouse file to upload
    directory_path: path to the local directory containing files to upload
    """

    settings = UploaderSettings(**kwargs)
    file_type = settings.file_type

    if file_type == FileType.DBT:
        assert not settings.directory_path
        upload_manifest(
            token=settings.token,
            source_id=settings.source_id,
            file_path=settings.file_path,
            zone=settings.zone,
        )
        return None

    upload(
        token=settings.token,
        source_id=settings.source_id,
        file_type=file_type,
        file_path=settings.file_path,
        directory_path=settings.directory_path,
        zone=settings.zone,
    )



import json
import logging
from dataclasses import dataclass
from datetime import datetime
from enum import Enum
from typing import Literal

import requests
from dateutil.parser import parse

from ...utils.url import add_path
from .credentials import DbtCredentials

logger = logging.getLogger(__name__)


_URL_SUFFIX = "/api/v2/accounts/"

_DATA_KEY = "data"
_SUCCESSFUL_RUN_STATUS = 10
_DELETED_JOB_STATE = 2

Artifact = Literal["manifest.json", "run_results.json"]


class ContentType(Enum):
    JSON = "application/json"
    TEXT = "text/html"


@dataclass(frozen=True)
class DbtRun:
    id: int
    finished_at: datetime


def _account_url(host: str) -> str:
    host = host.rstrip("/")
    return f"{host}{_URL_SUFFIX}"


def _is_deleted(job: dict) -> bool:
    return job["state"] == _DELETED_JOB_STATE


class DbtClient:
    """
    Connect to dbt-cloud Administrative API and downloads manifest.
    https://docs.getdbt.com/docs/dbt-cloud-apis/admin-cloud-api
    """

    def __init__(self, credentials: DbtCredentials):
        self._credentials = credentials
        self._account_url = _account_url(self._credentials.host)
        self._session = requests.Session()
        self._account_id: str = self._infer_account_id()

    def _headers(self, content_type: ContentType) -> dict:
        return {
            "Accept": content_type.value,
            "Authorization": "Token " + self._credentials.token,
        }

    def _call(
        self,
        url: str,
        params: dict | None = None,
        content_type: ContentType = ContentType.JSON,
        ignore_not_found: bool = False,
    ) -> dict:
        headers = self._headers(content_type)
        response = self._session.get(
            url=url,
            headers=headers,
            params=params,
        )

        if ignore_not_found and response.status_code == 404:
            logger.info(f"No results found at {url} (params: {params})")
            return {}

        try:
            result = response.json()
        except:
            context = f"{url}, status {response.status_code}"
            raise ValueError(f"Couldn't extract data from {context}")
        else:
            if content_type == ContentType.JSON:
                return result[_DATA_KEY]
            return result

    def _infer_account_id(self) -> str:
        result = self._call(url=self._account_url)
        return str(result[0]["id"])

    def list_job_identifiers(self) -> set[int]:
        """
        Return the IDs of all non-deleted jobs for this account
        https://docs.getdbt.com/dbt-cloud/api-v2-legacy#tag/Jobs/operation/listJobsForAccount
        """
        url = add_path(self._account_url, self._account_id, "jobs", "/")
        jobs = self._call(url)
        return {job["id"] for job in jobs if not _is_deleted(job)}

    def last_run(
        self,
        job_id: int | None = None,
        finished_at_range: tuple[datetime, datetime] | None = None,
    ) -> DbtRun | None:
        """
        Extract the last successful run id, optionally filtered on a given datetime range
        https://docs.getdbt.com/dbt-cloud/api-v2#tag/Runs/operation/listRunsForAccount
        """
        url = add_path(self._account_url, self._account_id, "runs", "/")

        params = {
            "job_definition_id": job_id or self._credentials.job_id,
            "order_by": "-finished_at",
            "status": _SUCCESSFUL_RUN_STATUS,
            "limit": 1,
        }

        if finished_at_range:
            # https://github.com/dbt-labs/dbt-cloud-openapi-spec/issues/33
            params["finished_at__range"] = json.dumps(
                [d.isoformat() for d in finished_at_range]
            )

        data = self._call(url, params)
        if not data:
            return None

        if not data[0]:
            raise ValueError("API result is empty")

        run = data[0]
        return DbtRun(id=run["id"], finished_at=parse(run["finished_at"]))

    def _fetch_run_artifact(
        self, run_id: int, artifact: Artifact, ignore_not_found=False
    ) -> dict | None:
        """
        Fetch dbt manifest or run results
        https://docs.getdbt.com/dbt-cloud/api-v2-legacy#tag/Runs/operation/getArtifactsByRunId
        """
        url = add_path(
            self._account_url,
            self._account_id,
            "runs",
            str(run_id),
            "artifacts",
            artifact,
        )
        logger.info(
            f"Extracting {artifact} from run id {run_id} with url {url}"
        )

        # setting text content as a workaround to this issue
        # https://stackoverflow.com/questions/68201659/dbt-cloud-api-to-extract-run-artifacts
        return self._call(
            url,
            content_type=ContentType.TEXT,
            ignore_not_found=ignore_not_found,
        )

    def fetch_manifest(self, run_id: int | None = None) -> dict:
        """
        Fetch the manifest or the given run, or of the latest run for the configured job
        """
        if not run_id:
            run = self.last_run()
            assert run, f"No run found for job {self._credentials.job_id}"
            run_id = run.id
        manifest = self._fetch_run_artifact(run_id, "manifest.json")
        assert manifest, f"No manifest found for run id {run_id}"
        return manifest

    def fetch_run_results(self, run_id: int) -> dict | None:
        """
        Fetch the run results of the given run.
        It seems the API doesn't always find the file, in which case None is returned.
        """
        return self._fetch_run_artifact(
            run_id, "run_results.json", ignore_not_found=True
        )



from dataclasses import dataclass
from typing import Any
from unittest.mock import Mock

import pytest
from requests.exceptions import ReadTimeout

from .client import COALESCE_MIN_PAGE_SIZE, COALESCE_PAGE_SIZE, CoalesceClient
from .credentials import CoalesceCredentials
from .endpoint import CoalesceEndpointFactory


@dataclass
class _StubResponse:
    payload: dict[str, Any]

    def json(self) -> dict[str, Any]:
        return self.payload


def _make_client() -> CoalesceClient:
    # Minimal real client instance; we'll patch _call so no network happens.
    creds = CoalesceCredentials(
        host="https://example.invalid", token="fake-token"
    )
    return CoalesceClient(credentials=creds)


def _node(i: int) -> dict[str, Any]:
    return {"id": f"{i:03d}", "name": f"n{i}"}


def test_fetch_env_nodes_paginates_until_next_is_none(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    """
    Ensures _fetch_env_nodes() keeps fetching pages using the 'next' cursor and
    stops when next is None.
    """
    client = _make_client()
    env_id = 17
    expected_endpoint = CoalesceEndpointFactory.nodes(environment_id=env_id)

    # Simulate 3 pages: 200 + 200 + 305 = 705
    pages: dict[str | None, _StubResponse] = {
        None: _StubResponse(
            {
                "data": [_node(i) for i in range(200)],
                "limit": 200,
                "next": "c1",
                "total": 200,
            }
        ),
        "c1": _StubResponse(
            {
                "data": [_node(i) for i in range(200, 400)],
                "limit": 200,
                "next": "c2",
                "total": 200,
            }
        ),
        "c2": _StubResponse(
            {
                "data": [_node(i) for i in range(400, 705)],
                "limit": 200,
                "next": None,
                "total": 305,
            }
        ),
    }

    def _call_stub(
        *, method: str, endpoint: str, params: dict, retry_on_timeout: bool
    ) -> _StubResponse:
        assert method == "GET"
        assert endpoint == expected_endpoint
        assert params.get("detail") == "true"
        cursor = params.get("startingFrom")
        return pages[cursor]

    call_mock = Mock(side_effect=_call_stub)
    monkeypatch.setattr(client, "_call", call_mock)

    nodes = client._fetch_env_nodes(env_id)

    assert len(nodes) == 705
    assert all(n["environment_id"] == env_id for n in nodes)
    assert call_mock.call_count == 3


def test_fetch_env_nodes_reduces_page_size_on_timeout(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    """
    If a timeout happens, smart pagination should reduce page_size and retry the SAME cursor.
    """
    client = _make_client()
    env_id = 17

    calls: list[dict[str, Any]] = []

    def _call_stub(
        *,
        method: str,
        endpoint: str,
        params: dict,
        retry_on_timeout: bool,
    ) -> _StubResponse:
        calls.append(params)
        assert params.get("detail") == "true"

        # First attempt fails at default page size.
        if len(calls) == 1:
            assert params.get("limit") == COALESCE_PAGE_SIZE
            raise ReadTimeout("boom")

        # Second attempt should be a retry with a reduced page size (still first page cursor).
        limit = params.get("limit")
        assert isinstance(limit, int)
        assert limit < COALESCE_PAGE_SIZE
        assert (
            params.get("startingFrom") is None or "startingFrom" not in params
        )

        # Return a single page and finish.
        return _StubResponse(
            {
                "data": [_node(0)],
                "limit": params["limit"],
                "next": None,
                "total": 1,
            }
        )

    call_mock = Mock(side_effect=_call_stub)
    monkeypatch.setattr(client, "_call", call_mock)

    # Avoid sleeping in tests
    monkeypatch.setattr("time.sleep", lambda *_args, **_kwargs: None)

    nodes = client._fetch_env_nodes(env_id)

    assert len(nodes) == 1
    assert call_mock.call_count == 2


def test_fetch_env_nodes_raises_when_page_size_is_1_and_still_timeouts(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    """
    Coalesce cursor pagination can't skip assets. When page_size is reduced to 1 and it still fails,
    we must raise.
    """
    client = _make_client()
    env_id = 17

    def _call_stub(
        *,
        method: str,
        endpoint: str,
        params: dict,
        retry_on_timeout: bool,
    ) -> _StubResponse:
        assert params.get("detail") == "true"
        raise ReadTimeout("still failing")

    call_mock = Mock(side_effect=_call_stub)
    monkeypatch.setattr(client, "_call", call_mock)
    monkeypatch.setattr("time.sleep", lambda *_args, **_kwargs: None)

    # Force starting page_size to MIN so it raises immediately on first error at page_size=1.
    monkeypatch.setattr(
        "castor_extractor.transformation.coalesce.client.client.COALESCE_PAGE_SIZE",
        COALESCE_MIN_PAGE_SIZE,
    )

    with pytest.raises(ReadTimeout):
        client._fetch_env_nodes(env_id)



from dataclasses import dataclass, field
from datetime import datetime
from enum import Enum


@dataclass
class QueryPart:
    """Query fragment from Redshift"""

    query_id: str
    text: str
    sequence: int
    sequence_count: int

    def __post_init__(self) -> None:
        if not isinstance(self.query_id, str):
            self.query_id = str(self.query_id)


@dataclass
class QueryBuffer:
    """Buffer to hold query parts that have yet to be reassembled"""

    expected: int
    parts: dict[int, str] = field(default_factory=dict)


@dataclass
class AssembledQuery:
    """Reassembled redshift query"""

    query_id: str
    text: str


@dataclass
class QueryMetadata:
    """Query/ddl metadata"""

    aborted: int
    database_id: int
    database_name: str
    end_time: datetime
    label: str
    process_id: str
    query_id: str
    start_time: datetime
    user_id: int
    user_name: str


class LongQuery(Enum):
    METADATA = "metadata"
    PARTS = "parts"



import pytest

from .properties import level_1_folder_id


def test_level_1_folder_id():
    folders = [
        {"id": "toto", "level": 2},
        {"id": "tata", "level": 1},
        {"id": "tutu", "level": 3},
    ]
    assert level_1_folder_id(folders) == "tata"

    folders = []
    with pytest.raises(ValueError, match="No level 1 folder found"):
        level_1_folder_id(folders)



from dataclasses import field

from pydantic_settings import BaseSettings, SettingsConfigDict

CASTOR_ENV_PREFIX = "CASTOR_GLUE_"


class GlueAthenaCredentials(BaseSettings):
    """
    Credentials needed by Glue/Athena client
    https://docs.coalesce.io/docs/catalog/integrations/data-warehouses/glue
    """

    access_key_id: str = field(metadata={"sensitive": True})
    access_key_secret: str = field(metadata={"sensitive": True})
    aws_account_id: str = field(metadata={"sensitive": True})
    aws_region: str

    model_config = SettingsConfigDict(
        env_prefix=CASTOR_ENV_PREFIX,
        extra="ignore",
        populate_by_name=True,
    )



import logging
import math

logger = logging.getLogger(__name__)


class SmartPagination:
    """
    Pagination helper that dynamically reduces page size when encountering
    slow or timing-out API responses.

    It supports progressive fallback
    (e.g., 200 → 100 → 10 → 1) and resets back to the initial page size once
    the problematic segment is passed.
    """

    def __init__(self, initial_page_size: int) -> None:
        self.initial_page_size = initial_page_size
        self._reduced_page_size = None
        self._counter: int | None = None
        self._slow_mode = False

    @property
    def page_size(self):
        return self._reduced_page_size or self.initial_page_size

    def reduce_page_size(self):
        new_page_size = math.ceil(self.page_size / 10)
        logger.info(
            f"Page size reduced from {self.page_size} to {new_page_size}"
        )
        self._reduced_page_size = new_page_size
        if not self._slow_mode:
            # we just entered a problematic batch
            # start the countdown
            self._counter = 0
            self._slow_mode = True

    def reset(self):
        """
        Manual reset is useful after skipping a faulty row: we want to get back
        at full speed, under the assumption that the timeouts are behind us
        """
        logger.info(f"Resetting initial page size: {self.initial_page_size}")
        self._reduced_page_size = None
        self._counter = None
        self._slow_mode = False

    def next(self):
        """
        Advance the counter when operating in slow mode.
        Once the problematic segment has been fully traversed,
        automatically restore the initial pagination strategy to resume
        efficient extraction.
        """
        if not self._slow_mode:
            return
        self._counter += self._reduced_page_size
        if self._counter >= self.initial_page_size:
            self.reset()



from typing import Literal

import msal  # type: ignore

from ....utils import BearerAuth
from .constants import Keys
from .credentials import PowerbiCertificate, PowerbiCredentials
from .endpoints import PowerBiEndpointFactory

TScope = Literal["powerbi", "graph"]


def _get_client_credential(
    secret: str | None, certificate: PowerbiCertificate | None
) -> str | dict:
    if secret:
        return secret
    if certificate:
        return certificate.model_dump()

    raise ValueError("Either certificate or secret must be provided.")


class PowerBiBearerAuth(BearerAuth):
    """
    Class to handle authentication to both the Power BI API and
    the Microsoft Graph API.
    The only difference lies in the scopes.
    """

    def __init__(
        self,
        credentials: PowerbiCredentials,
        scope_type: TScope = "powerbi",
    ):
        self.credentials = credentials
        self.scopes = self._pick_scopes(scope_type)

        endpoint_factory = PowerBiEndpointFactory(
            login_url=self.credentials.login_url,
            api_base=self.credentials.api_base,
        )
        authority = endpoint_factory.authority(self.credentials.tenant_id)

        client_credential = _get_client_credential(
            self.credentials.secret, self.credentials.certificate
        )

        self.app = msal.ConfidentialClientApplication(
            client_id=self.credentials.client_id,
            authority=authority,
            client_credential=client_credential,
        )

    def _pick_scopes(self, scope_type: TScope) -> list[str]:
        if scope_type == "powerbi":
            return self.credentials.powerbi_scopes

        if scope_type == "graph":
            return self.credentials.graph_scopes

        raise ValueError("scope_type must be either 'powerbi' or 'graph'")

    def fetch_token(self):
        token = self.app.acquire_token_for_client(scopes=self.scopes)

        if Keys.ACCESS_TOKEN not in token:
            raise ValueError(f"No access token in token response: {token}")

        return token[Keys.ACCESS_TOKEN]



from unittest import mock
from unittest.mock import MagicMock

import pytest

from .client import StrategyClient
from .credentials import StrategyCredentials
from .properties import (
    ObjectSubTypes,
    ObjectTypes,
    format_url,
)

_LOCATION = "Project/ABC123/Folder/My Document"
_NAME = "My Document"
_PROJECT_ID = "ABC123"

_CUBE_ID = "CUBE"
_DASHBOARD_ID = "DASH"
_DOCUMENT_ID = "DOC"
_FOLDER_ID = "FOLD"
_REPORT_ID = "REP"


@pytest.fixture
def mocked_cube():
    cube = MagicMock()
    cube.ancestors = [{"id": _FOLDER_ID, "level": 1}]
    cube.id = _CUBE_ID
    cube.project_id = _PROJECT_ID
    cube.subtype = ObjectSubTypes.OLAP_CUBE.value
    cube.type = ObjectTypes.REPORT_DEFINITION
    return cube


@pytest.fixture
def mocked_dashboard():
    dashboard = MagicMock()
    dashboard.id = _DASHBOARD_ID
    dashboard.project_id = _PROJECT_ID
    dashboard.type = ObjectTypes.DOCUMENT_DEFINITION
    return dashboard


@pytest.fixture
def mocked_document():
    document = MagicMock()
    document.id = _DOCUMENT_ID
    document.location = _LOCATION
    document.name = _NAME
    document.project_id = _PROJECT_ID
    document.type = ObjectTypes.DOCUMENT_DEFINITION
    return document


@pytest.fixture
def mocked_report():
    report = MagicMock()
    report.id = _REPORT_ID
    report.project_id = _PROJECT_ID
    report.type = ObjectTypes.REPORT_DEFINITION
    return report


def test_format_url(
    mocked_cube,
    mocked_dashboard,
    mocked_document,
    mocked_report,
):
    credentials = StrategyCredentials(
        base_url="https://reporting.catysserie.com/MicroStrategyLibrary",
        username="user",
        password="pass",
        project_ids=[_PROJECT_ID],
    )

    connection_path = "source.packages.extractor.castor_extractor.visualization.strategy.client.client.Connection"
    is_dashboard_path = "source.packages.extractor.castor_extractor.visualization.strategy.client.properties.is_dashboard"
    with (
        mock.patch(connection_path),
        mock.patch(is_dashboard_path) as mock_is_dashboard,
    ):
        client = StrategyClient(credentials)
        assert client.hostname == "reporting.catysserie.com"

        assert (
            format_url(mocked_cube, client.hostname)
            == "https://reporting.catysserie.com/MicroStrategy/servlet/mstrWeb?folderID=FOLD&projectID=ABC123"
        )

        mock_is_dashboard.return_value = True
        assert (
            format_url(mocked_dashboard, client.hostname)
            == "https://reporting.catysserie.com/MicroStrategyLibrary/app/ABC123/DASH"
        )

        mock_is_dashboard.return_value = False
        assert (
            format_url(mocked_document, client.hostname)
            == "https://reporting.catysserie.com/MicroStrategy/servlet/mstrWeb?documentID=DOC&projectID=ABC123"
        )

        assert (
            format_url(mocked_report, client.hostname)
            == "https://reporting.catysserie.com/MicroStrategy/servlet/mstrWeb?reportID=REP&projectID=ABC123"
        )


def test_StrategyClient__common_entity_properties(mocked_document):
    credentials = StrategyCredentials(
        base_url="https://reporting.catysserie.com/MicroStrategyLibrary",
        username="user",
        password="pass",
        project_ids=[_PROJECT_ID],
    )

    connection_path = "source.packages.extractor.castor_extractor.visualization.strategy.client.client.Connection"
    list_deps_path = "source.packages.extractor.castor_extractor.visualization.strategy.client.client.list_dependencies"
    safe_get_attr_path = "source.packages.extractor.castor_extractor.visualization.strategy.client.client.safe_get_property"
    url_path = "source.packages.extractor.castor_extractor.visualization.strategy.client.client.format_url"

    with (
        mock.patch(connection_path),
        mock.patch(list_deps_path) as mock_list_deps,
        mock.patch(safe_get_attr_path) as mock_safe_get_attr,
        mock.patch(url_path) as mock_url_path,
    ):
        client = StrategyClient(credentials)

        # mock url
        url = "https://some.url.com/document"
        mock_url_path.return_value = url

        # mock dependencies
        mock_list_deps.return_value = [
            {"id": "1", "name": "attr", "subtype": 666, "type": 12}
        ]

        # mock _safe_get_attribute
        description = "some description"
        mock_safe_get_attr.return_value = description

        properties = client._common_entity_properties(
            mocked_document, project_id=_PROJECT_ID
        )

        assert properties["description"] == description
        assert properties["id"] == _DOCUMENT_ID
        assert properties["location"] == _LOCATION
        assert properties["name"] == _NAME
        assert properties["type"] == ObjectTypes.DOCUMENT_DEFINITION.value
        assert properties["url"] == url
        assert properties["project_id"] == _PROJECT_ID

        mock_url_path.reset_mock()
        mock_safe_get_attr.reset_mock()

        properties_no_url_or_desc = client._common_entity_properties(
            mocked_document,
            project_id=_PROJECT_ID,
            with_url=False,
            with_description=False,
        )
        assert "url" not in properties_no_url_or_desc
        assert "description" not in properties_no_url_or_desc
        assert mock_url_path.call_count == 0
        assert mock_safe_get_attr.call_count == 0



import logging
from collections.abc import Iterable

from ...utils import (
    OUTPUT_DIR,
    current_timestamp,
    deep_serialize,
    from_env,
    get_output_filename,
    write_json,
    write_summary,
)
from .assets import StrategyAsset
from .client import StrategyClient, StrategyCredentials

logger = logging.getLogger(__name__)


def iterate_all_data(
    client: StrategyClient,
) -> Iterable[tuple[str, list | dict]]:
    """Iterate over the extracted data from Strategy"""

    for asset in StrategyAsset.mandatory:
        logger.info(f"Extracting {asset.value.upper()} from REST API")
        data = client.fetch(asset)
        yield asset.name.lower(), list(deep_serialize(data))


def extract_all(**kwargs) -> None:
    _output_directory = kwargs.get("output") or from_env(OUTPUT_DIR)
    credentials = StrategyCredentials(**kwargs)

    client = StrategyClient(credentials=credentials)
    ts = current_timestamp()

    for key, data in iterate_all_data(client):
        filename = get_output_filename(key, _output_directory, ts)
        write_json(filename, data)

    client.close()
    write_summary(_output_directory, ts)



from dataclasses import dataclass

Params = dict[str, str]  # key: value


def values_to_params(prefix: str, values: list[str] | None) -> Params:
    """
    Convert a list of values to a dictionary.
    Each value is assigned a key based on the given prefix and its
    position in the list
    """
    if not values:
        return {}
    return {f"{prefix}_{i}": value for i, value in enumerate(values)}


@dataclass
class QueryFilter:
    """
    Represent a query filter, composed of:
    - The name of the {placeholder} that should be replaced in the query
    - The SQL expression or fragment to replace the placeholder with
    - Optional parameters used for value binding during query execution.

    Example:
    > SELECT ... FROM users WHERE {user_blocked}

    filter = QueryFilter(
        placeholder = "user_blocked",
        expression = "user_name NOT IN (user_1:, user_2:)",
        params = {"user_1:": "John", "user_2:": "Mary"}
    )
    """

    placeholder: str
    expression: str
    params: Params | None = None


class ExtractionQuery:
    """
    Contains useful context to run the query:
    - the sql statement itself
    - parameters { ... }
    - optionally, the target database (can be used to change the engine's URI)
    """

    def __init__(
        self,
        statement: str,
        params: dict,
        database: str | None = None,
    ):
        self.statement = statement
        self.params = params
        self.database = database

    def apply_filters(self, filters: list[QueryFilter]) -> None:
        """
        Apply a list of filters to this query.

        Each filter replaces a corresponding placeholder in the statement
        and updates the query's parameters with any additional filter-specific
        parameters.
        """
        placeholder_expressions = {f.placeholder: f.expression for f in filters}
        self.statement = self.statement.format(**placeholder_expressions)
        for f in filters:
            self.params.update(f.params or {})



from ..abstract import WarehouseAsset
from .query import INCOMPATIBLE_WITH_OMNI, _omni_compatible


def test__omni_compatible():
    # asset that is incompatible with omni
    asset = next(iter(INCOMPATIBLE_WITH_OMNI))

    regions = {
        ("p1", "us-central1"),
        ("p1", "aws-us-east-1"),
        ("p1", "azure-europe-west3"),
        ("p2", "europe-west1"),
    }

    filtered = _omni_compatible(regions, asset)

    # non-omni locations should remain
    assert ("p1", "us-central1") in filtered
    assert ("p2", "europe-west1") in filtered

    # omni locations must be removed
    assert ("p1", "aws-us-east-1") not in filtered
    assert ("p1", "azure-europe-west3") not in filtered

    # pick an asset that is NOT in the incompatible list
    compatible_asset = next(
        a for a in WarehouseAsset if a not in INCOMPATIBLE_WITH_OMNI
    )

    regions = {
        ("p1", "us-central1"),
        ("p1", "aws-us-east-1"),
        ("p1", "azure-europe-west3"),
    }

    filtered = _omni_compatible(regions, compatible_asset)

    # nothing should be filtered out
    assert filtered == regions



from datetime import datetime

from ..abstract import TimeFilter, WarehouseAsset
from .query_builder import SnowflakeQueryBuilder


def test_snowflake_query_builder():
    day = datetime(1969, 7, 20)
    time_filter = TimeFilter(day=day, hour_min=20, hour_max=21)

    # Step 1 - empty filters
    builder = SnowflakeQueryBuilder(
        time_filter=time_filter,
        db_allowed=None,
        db_blocked=None,
        query_blocked=None,
    )
    query = builder.build(WarehouseAsset.QUERY)[0]
    assert len(query.params) == 3  # day, hour_min, hour_max
    assert "{query_blocked}" not in query.statement
    # When db_allowed, db_blocked, and query_blocked are all None,
    # they're all replaced with "TRUE" (3 total)
    assert query.statement.count("AND TRUE") == 3

    query = builder.build(WarehouseAsset.SCHEMA)[0]
    assert len(query.params) == 0
    assert "{db_allowed}" not in query.statement
    assert "{db_blocked}" not in query.statement
    # For SCHEMA, db_allowed and db_blocked are replaced with "TRUE" (2 total)
    assert query.statement.count("AND TRUE") == 2

    # Step 2 - filters
    builder = SnowflakeQueryBuilder(
        time_filter=time_filter,
        db_allowed=["foo", "bar"],
        db_blocked=None,
        query_blocked=["SHOW DATABASES", "SELECT 1 FROM %"],
    )
    query = builder.build(WarehouseAsset.QUERY)[0]
    assert len(query.params) == 5  # 3 time filters + 2 query filters
    assert "{query_blocked}" not in query.statement
    assert query.params["query_blocked_0"] == "SHOW DATABASES"
    assert query.params["query_blocked_1"] == "SELECT 1 FROM %"
    expected_chunk = "query_text NOT ILIKE :query_blocked"
    assert f"AND ({expected_chunk}_0 AND {expected_chunk}_1)" in query.statement

    query = builder.build(WarehouseAsset.SCHEMA)[0]
    assert len(query.params) == 2  # 2 database filters
    assert "{database_allowed}" not in query.statement
    assert "{database_blocked}" not in query.statement
    assert query.params["database_allowed_0"] == "foo"
    assert query.params["database_allowed_1"] == "bar"
    expected = "database_name IN (:database_allowed_0, :database_allowed_1)"
    assert expected in query.statement

    # Step 3 - apply_db_filter_to_queries=False disables db filtering for QUERY assets
    builder = SnowflakeQueryBuilder(
        time_filter=time_filter,
        db_allowed=["foo", "bar"],
        db_blocked=["baz"],
        apply_db_filter_to_queries=False,
        query_blocked=None,
    )
    query = builder.build(WarehouseAsset.QUERY)[0]
    # only time filters, db filters replaced with TRUE
    assert len(query.params) == 3
    # query_blocked + db_allowed + db_blocked all TRUE
    assert query.statement.count("AND TRUE") == 3

    # Step 4 - apply_db_filter_to_queries=False doesn't affect SCHEMA assets
    query = builder.build(WarehouseAsset.SCHEMA)[0]
    assert len(query.params) == 3  # 2 db_allowed + 1 db_blocked
    assert (
        "database_name IN (:database_allowed_0, :database_allowed_1)"
        in query.statement
    )

    # Step 5 - apply_db_filter_to_queries=True applies db filtering to QUERY assets
    builder = SnowflakeQueryBuilder(
        time_filter=time_filter,
        db_allowed=["foo", "bar"],
        db_blocked=["baz"],
        apply_db_filter_to_queries=True,
        query_blocked=None,
    )
    query = builder.build(WarehouseAsset.QUERY)[0]
    # 3 time filters + 3 db filters (2 allowed + 1 blocked)
    assert len(query.params) == 6
    assert (
        "database_name IN (:database_allowed_0, :database_allowed_1)"
        in query.statement
    )
    assert "database_name NOT IN (:database_blocked_0)" in query.statement



from typing import Optional

from pydantic import Field
from pydantic_settings import BaseSettings, SettingsConfigDict

CASTOR_ENV_PREFIX = "CASTOR_SALESFORCE_"


class SalesforceCredentials(BaseSettings):
    """
    Class to handle Salesforce rest API permissions
    """

    model_config = SettingsConfigDict(
        env_prefix=CASTOR_ENV_PREFIX,
        extra="ignore",
        populate_by_name=True,
    )

    base_url: str
    client_id: str
    client_secret: str = Field(repr=False)
    password: Optional[str] = Field(repr=False, default=None)
    security_token: Optional[str] = Field(repr=False, default=None)
    username: Optional[str] = None

    @property
    def password_token(self) -> str:
        """Generates the password for authentication"""
        assert self.password  # for mypy
        assert self.security_token  # for mypy
        return self.password + self.security_token

    def token_request_payload(self) -> dict[str, str]:
        """
        Params to post to the API in order to retrieve the authentication token
        """
        if all(
            v is None
            for v in (self.username, self.password, self.security_token)
        ):
            return {
                "grant_type": "client_credentials",
                "client_id": self.client_id,
                "client_secret": self.client_secret,
            }

        for key in ("username", "password", "security_token"):
            if getattr(self, key) is None:
                raise ValueError(
                    "'{}' was not provided. It is required for Basic Authentication. If you intend to use Credential Client Flow, make sure not to provide any of: \
                    username, password, security_token".format(key)
                )
        assert self.username  # for mypy
        assert self.password  # for mypy
        assert self.security_token  # for mypy
        return {
            "grant_type": "password",
            "client_id": self.client_id,
            "client_secret": self.client_secret,
            "username": self.username,
            "password": self.password_token,
        }



import logging
from collections.abc import Iterable

from ...utils import (
    OUTPUT_DIR,
    current_timestamp,
    deep_serialize,
    from_env,
    get_output_filename,
    write_json,
    write_summary,
)
from .assets import QlikAsset
from .client import QlikClient, QlikCredentials

logger = logging.getLogger(__name__)


def iterate_all_data(
    client: QlikClient,
    include_sheets: bool | None = False,
) -> Iterable[tuple[QlikAsset, list | dict]]:
    """Iterate over the extracted data from Qlik"""

    logger.info("Extracting CONNECTIONS from REST API")
    connections = client.fetch(QlikAsset.CONNECTIONS)
    yield QlikAsset.CONNECTIONS, deep_serialize(connections)

    logger.info("Extracting SPACES from REST API")
    spaces = client.fetch(QlikAsset.SPACES)
    yield QlikAsset.SPACES, deep_serialize(spaces)

    logger.info("Extracting USERS from REST API")
    users = client.fetch(QlikAsset.USERS)
    yield QlikAsset.USERS, deep_serialize(users)

    logger.info("Extracting APPS from REST API")
    apps = client.fetch(QlikAsset.APPS)
    yield QlikAsset.APPS, deep_serialize(apps)

    logging.info("Extracting LINEAGE data from REST API")
    lineage = client.fetch(QlikAsset.LINEAGE, apps=apps)
    yield QlikAsset.LINEAGE, deep_serialize(lineage)

    logging.info("Extracting MEASURES data from JSON-RPC API")
    measures = client.fetch(QlikAsset.MEASURES, apps=apps)
    yield QlikAsset.MEASURES, deep_serialize(measures)

    if include_sheets:
        logger.info("Extracting SHEETS from JSON-RPC API")
        sheets = client.fetch(QlikAsset.SHEETS, apps=apps)
        yield QlikAsset.SHEETS, deep_serialize(sheets)


def extract_all(
    except_http_error_statuses: list[int] | None = None, **kwargs
) -> None:
    """
    Extract data from Qlik REST API
    Store the output files locally under the given output_directory
    """

    credentials = QlikCredentials(**kwargs)
    _output_directory = kwargs.get("output") or from_env(OUTPUT_DIR)
    include_sheets = kwargs.get("include_sheets")

    client = QlikClient(
        credentials=credentials,
        except_http_error_statuses=except_http_error_statuses,
    )

    ts = current_timestamp()

    for key, data in iterate_all_data(client, include_sheets):
        filename = get_output_filename(key.name.lower(), _output_directory, ts)
        write_json(filename, data)

    write_summary(_output_directory, ts, base_url=credentials.base_url)



import logging

from ...utils import ExtractionQuery
from ..abstract import (
    AbstractQueryBuilder,
    TimeFilter,
    WarehouseAsset,
)

logger = logging.getLogger(__name__)


_NO_DATABASE_ERROR_MSG = (
    "No databases eligible for extraction. "
    "If you are using the db_allow/db_block options, please make sure to use the correct case."
)

_SQLSERVER_DUPLICATED_ASSETS: tuple[WarehouseAsset, ...] = (
    WarehouseAsset.USER,
)


class MSSQLQueryBuilder(AbstractQueryBuilder):
    """
    Builds queries to extract assets from SQL Server.
    """

    def __init__(
        self,
        databases: list[str],
        query_databases: list[str],
        time_filter: TimeFilter | None = None,
    ):
        super().__init__(
            time_filter=time_filter,
            duplicated=_SQLSERVER_DUPLICATED_ASSETS,
        )
        if not databases:
            raise ValueError(_NO_DATABASE_ERROR_MSG)
        self._databases = databases
        self._query_databases = query_databases

    @staticmethod
    def _format(
        query: ExtractionQuery,
        database: str,
    ) -> ExtractionQuery:
        return ExtractionQuery(
            statement=query.statement.format(database=database),
            params=query.params,
            database=database,
        )

    def build(self, asset: WarehouseAsset) -> list[ExtractionQuery]:
        query = self.build_default(asset)

        if asset == WarehouseAsset.DATABASE:
            # database.sql does not include a {database} placeholder.
            # Indeed, databases are extracted at the server level
            # (not scoped to a specific database).
            return [query]
        databases = (
            self._query_databases
            if asset == WarehouseAsset.QUERY
            else self._databases
        )
        logger.info(
            f"\tWill run queries with following database params: {databases}",
        )
        return [self._format(query, database) for database in databases]



import logging
from collections.abc import Iterator
from itertools import chain

from .client import PowerbiClient

logger = logging.getLogger(__name__)

_TYPE_USER = "User"
_TYPE_GROUP = "Group"


class UserDataProcessor:
    def __init__(self):
        self.users: list[dict] = []
        self.group_ids: set[str] = set()

    def parse_metadata(self, metadata: list[dict]) -> None:
        """
        Parses a list of individual users and a set of group IDs from
        the metadata. The list of users may contain duplicates.
        """
        self.users = []
        self.group_ids = set()

        for workspace in metadata:
            for user in workspace.get("users") or []:
                if user["principalType"] == _TYPE_USER:
                    self.users.append(user)

                elif user["principalType"] == _TYPE_GROUP:
                    self.group_ids.add(user["graphId"])

    @staticmethod
    def _normalize_group_member(group_member: dict) -> dict:
        """Normalizes a group member to match the Power BI user format."""
        return {
            "graphId": group_member["id"],
            "emailAddress": group_member["mail"],
            "displayName": group_member["displayName"],
        }

    def combine_users(self, group_members: list[dict]) -> Iterator[dict]:
        """
        Merges the given group members with the list of individual users found
        in the metadata, removing duplicates.
        """
        seen_users = set()
        normalized_group_members = map(
            self._normalize_group_member, group_members
        )

        for user in chain(self.users, normalized_group_members):
            id_ = user["graphId"]
            if id_ in seen_users:
                continue

            seen_users.add(id_)
            yield user


def consolidate_powerbi_users(
    client: PowerbiClient, metadata: list[dict]
) -> Iterator[dict]:
    """
    Extracts Power BI users as follows:
    - Metadata contains user information (individuals and groups)
    - Groups are expanded into its individual members
    - All users are consolidated into a single file

    If the group expansion fails due to a lack of permissions, it will raise
    a custom MicrosoftGraphAccessForbidden Exception.
    """
    user_processor = UserDataProcessor()
    user_processor.parse_metadata(metadata)

    logger.info("Extracting group members from Microsoft Graph API")
    group_members = list(
        client.fetch_group_members(group_ids=user_processor.group_ids)
    )
    combined_users = user_processor.combine_users(group_members)
    return combined_users



import functools
import itertools
import json
import logging
from collections.abc import Iterator

import requests
import tableauserverclient as TSC  # type: ignore
from tableauserverclient.server.endpoint.exceptions import (
    InternalServerError,
    NonXMLResponseError,
)

from ....utils import SerializedAsset, batch_of_length
from ....utils.client.api.smart_pagination import SmartPagination
from ..assets import TableauAsset
from ..constants import DEFAULT_PAGE_SIZE
from .errors import TableauApiError, TableauApiTimeout
from .gql_queries import FIELDS_QUERIES, GQL_QUERIES, QUERY_TEMPLATE

logger = logging.getLogger(__name__)

# increase the value when extraction is too slow
# decrease the value when timeouts arise
_CUSTOM_PAGE_SIZE: dict[TableauAsset, int] = {
    # fields are light but volumes are bigger
    TableauAsset.FIELD: 1000,
    # tables are sometimes heavy
    TableauAsset.TABLE: 50,
}

_TIMEOUT_MESSAGES = (
    "Execution canceled because timeout of 30000 millis was reached",
    # From experience, this internal error actually corresponds to a timeout
    "Internal Server Error(s) while executing query",
)


_BATCH_SIZE = 1000


def _format_filter_ids(filter_ids: list[str]) -> str:
    # "id_1", "id_2", ...
    formatted_filter_ids = ", ".join([f'"{id_}"' for id_ in filter_ids])

    # { idWithin: ["id_1", "id_2", ...] }
    return f"{{ idWithin: [{formatted_filter_ids}] }}"


def _is_timeout(error: dict) -> bool:
    error_message = error.get("message")
    return error_message in _TIMEOUT_MESSAGES


def _is_warning(error: dict) -> bool:
    extensions = error.get("extensions")
    if not extensions:
        return False

    severity = extensions.get("severity")
    if not severity:
        return False

    return severity.lower() == "warning"


def _check_errors(answer: dict) -> None:
    """
    Handle errors in graphql response:
    - return None when there's no errors in the answer
    - raise TableauApiTimeout if any of the errors is a timeout
    - else raise TableauApiError if any of the errors is critical
    - return None otherwise
    More info about Tableau errors:
    https://help.tableau.com/current/api/metadata_api/en-us/docs/meta_api_errors.html#other-errors
    """
    if "errors" not in answer:
        return

    errors = answer["errors"]

    has_timeout_errors = False
    has_critical_errors = False

    for error in errors:
        if _is_timeout(error):
            has_timeout_errors = True
            continue

        if _is_warning(error):
            # in this case, the answer contains the data anyway
            # just display the warning
            logger.warning(error)
            continue

        # at this point, it's not a timeout error
        # besides, it's not a warning (severity is either "error" or Unknown)
        has_critical_errors = True

    if has_timeout_errors:
        raise TableauApiTimeout(errors)
    if has_critical_errors:
        raise TableauApiError(errors)

    return None


_RETRYABLE_ERRORS = (
    InternalServerError,
    NonXMLResponseError,
    TableauApiTimeout,
    requests.exceptions.ReadTimeout,
)


def _is_relationship_service_war_exception(exception: Exception) -> bool:
    """
    Detects a specific NonXMLResponseError that occurs when Tableau's
    relationship-service-war endpoint responds with a 401 error due to
    concurrent API calls using the same authentication token.

    This issue is a known Tableau bug: when two metadata API calls happen
    simultaneously with the same token, Tableau may return a JSON payload
    instead of XML, wrapped in a NonXMLResponseError. Unlike other
    NonXMLResponseError cases, this one should **not** be retried — retrying
    will loop indefinitely and hide the real underlying token issue. Instead,
    callers should surface the exception and show logs to the user.

    More context:
    https://commtableau.my.site.com/s/question/0D54T00000dXBeRSAW/bug-report-unable-to-reauthenticate-once-the-token-expires

    """
    if not isinstance(exception, NonXMLResponseError):
        return False

    # raw bytes: b'{"timestamp":...}'
    error_bytes = exception.args[0]

    # Decode bytes to string
    error_str = error_bytes.decode("utf-8")
    # transform to JSON
    error_json = json.loads(error_str)

    status = error_json.get("status") or ""
    path = error_json.get("path") or ""

    return status == 401 and path == "/relationship-service-war/graphql"


def _gql_query_scroll(
    server,
    resource: str,
    fields: str,
    filter_: str,
    page_size: int,
    show_progress: bool,
) -> Iterator[SerializedAsset]:
    """
    Iterate over GQL query results, handling pagination and cursor

    We have a retry strategy when timeout issues arise.
    It's a known issue on Tableau side, still waiting for their fix:
    https://issues.salesforce.com/issue/a028c00000zKahoAAC/undefined
    """

    def _call(first: int, offset: int) -> dict:
        query = QUERY_TEMPLATE.format(
            resource=resource,
            fields=fields,
            filter=filter_,
            first=first,
            offset=offset,
        )
        answer = server.metadata.query(query)
        _check_errors(answer)
        return answer["data"][f"{resource}Connection"]

    current_offset = 0
    skipped_count = 0
    smart_pagination = SmartPagination(initial_page_size=page_size)
    while True:
        try:
            payload = _call(
                first=smart_pagination.page_size,
                offset=current_offset,
            )
            yield payload["nodes"]
            smart_pagination.next()

            current_offset += len(payload["nodes"])
            if show_progress:
                total = payload["totalCount"]
                logger.info(f"Extracted {current_offset}/{total} {resource}")

            if not payload["pageInfo"]["hasNextPage"]:
                break
        except _RETRYABLE_ERRORS as exception:
            if _is_relationship_service_war_exception(exception):
                raise
            logger.warning(f"Retryable exception: {exception}")
            if smart_pagination.page_size > 1:
                smart_pagination.reduce_page_size()
                continue

            logger.warning("Skipping asset because of TableauServer Timeout")
            skipped_count += 1
            current_offset += 1
            smart_pagination.reset()

    if skipped_count > 0:
        logger.info(f"Partial extraction - skipped {skipped_count} rows")


def _deduplicate(result_pages: Iterator[SerializedAsset]) -> SerializedAsset:
    """
    Sometimes assets are duplicated, which triggers UniqueViolation errors
    during store_all down the line.

    We suspect the offset pagination to be the root cause, because we had no
    problem until recently, when we switched from cursor pagination to offset
    pagination (for performance reasons)
    https://help.tableau.com/current/api/metadata_api/en-us/docs/meta_api_examples.html#pagination

    This is a straightforward solution to remove these duplicates directly at
    extraction.
    We don't show warnings because duplicates are expected, and we keep only
    the first occurrence since those duplicates are probably identical.
    """
    deduplicated: SerializedAsset = []
    seen_ids: set[str] = set()
    for page in result_pages:
        for asset in page:
            asset_id = asset["id"]
            if asset_id in seen_ids:
                # skip duplicate
                continue
            deduplicated.append(asset)
            seen_ids.add(asset_id)
    return deduplicated


class TableauClientMetadataApi:
    """
    Calls the MetadataAPI, using graphQL
    https://help.tableau.com/current/api/metadata_api/en-us/reference/index.html
    """

    def __init__(
        self,
        server: TSC.Server,
        override_page_size: int | None = None,
    ):
        self._server = server
        self._override_page_size = override_page_size

    def call(
        self,
        resource: str,
        fields: str,
        page_size: int = DEFAULT_PAGE_SIZE,
        show_progress: bool = True,
        filter_ids: list[str] | None = None,
    ) -> SerializedAsset:
        """
        Executes a GraphQL query against Tableau metadata API and returns
        deduplicated results.

        This method wraps the low-level pagination logic with retry handling,
        timeout management, optional batch skipping for error-prone assets,
        and post-processing to remove duplicate assets caused by offset
        pagination.

        This method handles two layers of chunking:
        1. ID-based batching: when `filter_ids` is provided, it is automatically
           split into batches of 1,000 to avoid query-size limits.
           Each batch is queried independently, then merged.

        2. Offset pagination: each batch (or the full resource when no
           `filter_ids` are given) is paginated using `page_size` via
           _gql_query_scroll()

        Results from all pages are concatenated and deduplicated.
        """

        scroll = functools.partial(
            _gql_query_scroll,
            server=self._server,
            resource=resource,
            fields=fields,
            page_size=page_size,
            show_progress=show_progress,
        )
        if filter_ids is None:
            # no ids –> run a single paginated query with an empty filter
            data = scroll(filter_="{}")
        else:
            # When IDs are provided:
            #   - Split IDs into manageable batches
            #   - Convert each batch into a GraphQL filter string
            #   - Query each batch independently
            filter_batches = (
                _format_filter_ids(ids)
                for ids in batch_of_length(filter_ids, _BATCH_SIZE)
            )
            data = itertools.chain.from_iterable(
                # Chain results from all batches.
                # Pagination progress is suppressed because it becomes too noisy
                # batching already emits sufficient logging.
                scroll(filter_=f, show_progress=False)
                for f in filter_batches
            )

        return _deduplicate(data)

    def _page_size(self, asset: TableauAsset) -> int:
        return (
            self._override_page_size
            or _CUSTOM_PAGE_SIZE.get(asset)
            or DEFAULT_PAGE_SIZE
        )

    def _fetch_fields(self) -> SerializedAsset:
        result: SerializedAsset = []
        page_size = self._page_size(TableauAsset.FIELD)
        for resource, fields in FIELDS_QUERIES:
            current = self.call(
                resource,
                fields,
                page_size,
            )
            result.extend(current)
        return result

    def fetch(
        self,
        asset: TableauAsset,
        filter_ids: list[str] | None = None,
    ) -> SerializedAsset:
        if asset == TableauAsset.FIELD:
            return self._fetch_fields()

        page_size = self._page_size(asset)
        resource, fields = GQL_QUERIES[asset]
        return self.call(
            resource=resource,
            fields=fields,
            page_size=page_size,
            filter_ids=filter_ids,
        )



from http import HTTPStatus
from statistics import variance
from time import time
from unittest.mock import patch

import pytest
import requests
from pydantic.error_wrappers import ValidationError
from requests import HTTPError, Response

from .retry import MS_IN_SEC, Retry, RetryStrategy, retry_request


def test_retry_field_validations():
    valid_args = {"max_retries": 1, "base_ms": 1000, "jitter_ms": 500}
    Retry(**valid_args)  # happy path

    failing_args_list = [
        {"max_retries": 0},
        {"base_ms": -523},
        {"jitter_ms": None},
        {"strategy": None},
    ]
    for failing_args in failing_args_list:
        args = {**valid_args, **failing_args}

        with pytest.raises(ValidationError):
            Retry(**args)


def _within(value: int, min_: int, max_: int) -> bool:
    return value >= min_ and value <= max_


def test_retry_strategy__jitter():
    retry = Retry(max_retries=2, base_ms=1000, jitter_ms=500)
    jitters = [retry.jitter() for _ in range(5)]

    # boundaries
    assert all(_within(j, 250, 750) for j in jitters)
    # randomness
    assert variance(jitters) > 0


def _iterate_base(retry: Retry, count: int) -> list[int]:
    bases: list[int] = []
    for _ in range(count):
        retry._retry_attempts += 1
        bases.append(retry.base())
    return bases


def test_retry_strategy__base():
    common_args = {"max_retries": 2, "base_ms": 2000, "jitter_ms": 500}
    # default strategy is constant
    retry = Retry(**common_args)
    assert _iterate_base(retry, 3) == [2000, 2000, 2000]
    retry = Retry(**common_args, strategy=RetryStrategy.CONSTANT)
    assert _iterate_base(retry, 3) == [2000, 2000, 2000]
    # linear
    retry = Retry(**common_args, strategy=RetryStrategy.LINEAR)
    assert _iterate_base(retry, 3) == [2000, 4000, 6000]
    # exponential
    # Set base_ms = 1000 to verify correct exponential growth.
    # With this value, (base_ms / 1000) == 1, so earlier implementations
    # incorrectly produced constant delays. This test ensures the fixed logic
    # now yields 1000, 2000, 4000, 8000 as expected.
    common_args["base_ms"] = 1000
    retry = Retry(**common_args, strategy=RetryStrategy.EXPONENTIAL)
    assert _iterate_base(retry, 4) == [1000, 2000, 4000, 8000]


def test_retry_strategy__check():
    retry = Retry(max_retries=3, base_ms=100, jitter_ms=10)
    error = ValueError

    before = time()
    assert retry.check(error)
    assert retry.check(error)
    assert retry.check(error)
    after = time()

    assert retry.check(error) is False
    delta_ms = int((after - before) * MS_IN_SEC)
    assert _within(delta_ms, 315, 345)


@patch("requests.get")
def test_retry_request(mocked_get):
    def error_response():
        response = Response()
        response.status_code = HTTPStatus.UNAUTHORIZED
        return response

    mocked_get.return_value = error_response()

    @retry_request(status_codes=(HTTPStatus.UNAUTHORIZED,), max_retries=3)
    def get():
        response = requests.get("hello")
        response.raise_for_status()
        return response.json()

    with pytest.raises(HTTPError):
        get()

    assert mocked_get.call_count == 4  # 1 call + 3 retries



from unittest.mock import Mock

import pytest

from .client import LookerStudioCredentials
from .parameters import (
    _BIGQUERY_CREDENTIALS_REQUIRED_ERROR,
    _EMPTY_USERS_FILE_ERROR,
    _LOOKER_STUDIO_CREDENTIALS_REQUIRED_ERROR,
    _USER_EMAIL_NOT_STRING_ERROR,
    _USERS_FILE_NOT_A_LIST_ERROR,
    ExtractionParameters,
)


def test_ExtractionParameters__validate_user_emails():
    base_params = {
        "bigquery_credentials": {"dummy": "credentials"},
        "has_source_queries_only": True,
        "has_view_activity_logs": False,
        "looker_studio_credentials": None,
        "output_directory": "tmp",
    }

    with pytest.raises(TypeError, match=_USERS_FILE_NOT_A_LIST_ERROR):
        ExtractionParameters(**base_params, user_emails="test")

    with pytest.raises(TypeError, match=_USERS_FILE_NOT_A_LIST_ERROR):
        ExtractionParameters(
            **base_params, user_emails={"not": "the", "right": "format"}
        )

    with pytest.raises(ValueError, match=_EMPTY_USERS_FILE_ERROR):
        ExtractionParameters(**base_params, user_emails=[])

    # Test non-string items
    with pytest.raises(TypeError, match=_USER_EMAIL_NOT_STRING_ERROR):
        ExtractionParameters(**base_params, user_emails=[1, 2, 3])

    # happy cases
    params = ExtractionParameters(
        **base_params, user_emails=["admin@toto.com", "tata@toto.com"]
    )
    assert params.user_emails == ["admin@toto.com", "tata@toto.com"]

    params = ExtractionParameters(**base_params, user_emails=None)
    assert params.user_emails is None


def test_ExtractionParameters_validate_credentials_requirements():
    base_params = {
        "has_view_activity_logs": False,
        "output_directory": "tmp",
        "user_emails": None,
    }

    # source queries only mode requires BigQuery credentials
    with pytest.raises(ValueError, match=_BIGQUERY_CREDENTIALS_REQUIRED_ERROR):
        ExtractionParameters(**base_params, has_source_queries_only=True)

    # happy path : source queries mode + BigQuery credentials
    ExtractionParameters(
        **base_params,
        has_source_queries_only=True,
        bigquery_credentials={"dummy": "credentials"},
    )

    # "normal" mode requires Looker Studio credentials
    with pytest.raises(
        ValueError,
        match=_LOOKER_STUDIO_CREDENTIALS_REQUIRED_ERROR,
    ):
        ExtractionParameters(
            **base_params,
            has_source_queries_only=False,
            looker_studio_credentials=None,
        )

    # happy path : source queries mode + both credentials
    ExtractionParameters(
        **base_params,
        has_source_queries_only=True,
        bigquery_credentials={"dummy": "credentials"},
        looker_studio_credentials=Mock(spec=LookerStudioCredentials),
    )



import logging
from collections.abc import Iterator
from dataclasses import asdict
from typing import Any

from ....utils import load_file
from ....warehouse.abstract import TimeFilter
from ....warehouse.bigquery import BigQueryClient
from ..assets import (
    CountAsset,
)
from .credentials import CountCredentials

logger = logging.getLogger(__name__)

_QUERIES_FOLDER = "queries"


class CountClient(BigQueryClient):
    """
    Count.co does not currently provide an official API.
    Instead, metadata such as dashboards, users, and queries is made available through
    special metadata tables stored in BigQuery.

    This client extends `BigQueryClient` to access and interact with those metadata tables.
    """

    def __init__(self, credentials: CountCredentials):
        super().__init__(asdict(credentials))
        self.project_id = credentials.project_id
        self.dataset_id = credentials.dataset_id
        self.time_filter = TimeFilter.default()  # setting current date - 1

    def _load_query(self, asset: CountAsset) -> str:
        query = load_file(
            f"{_QUERIES_FOLDER}/{asset.name.lower()}.sql", __file__
        )
        return query.format(
            project_id=self.project_id,
            dataset_id=self.dataset_id,
            extract_date=self.time_filter.day,
        )

    def fetch(self, asset: CountAsset) -> Iterator[dict[str, Any]]:
        """
        Fetch the asset given as a param by running a BigQuery query.
        """
        logger.info(f"Running BigQuery query to fetch: {asset.name}")

        query_str = self._load_query(asset)
        job = self.client.query(query_str)
        results = job.result()

        for row in results:
            yield dict(row)



import pytest

from .credentials import StrategyCredentials


def test_StrategyCredentials_project_ids():
    creds = {
        "base_url": "https://demo.strategy.com/MicroStrategyLibrary/api",
        "username": "castor",
        "password": "ilovepotatoes",
        "visualization_type": "strategy",
    }

    # no project IDs
    StrategyCredentials(**creds)

    # list of project IDs
    project_ids = ["ID1", "ID2", "ID3"]
    creds_using_list = {**creds, "project_ids": project_ids}
    credentials = StrategyCredentials(**creds_using_list)
    assert credentials.project_ids == project_ids

    # list of comma-separated of project IDs
    creds_using_comma_separated_ids = {**creds, "project_ids": " ID1 ,ID2, ID3"}
    credentials = StrategyCredentials(**creds_using_comma_separated_ids)
    assert credentials.project_ids == project_ids

    # single project ID string
    creds_using_single_id_string = {**creds, "project_ids": "ID1"}
    credentials = StrategyCredentials(**creds_using_single_id_string)
    assert credentials.project_ids == ["ID1"]

    with pytest.raises(ValueError):
        creds_with_unsupported_type = {**creds, "project_ids": 123456}
        StrategyCredentials(**creds_with_unsupported_type)



import logging
from collections.abc import Iterable, Iterator
from datetime import date

from ..assets import PowerBiAsset
from .credentials import PowerbiCredentials
from .graph_api_client import MicrosoftGraphPIClient
from .powerbi_api_client import PowerBIAPIClient

logger = logging.getLogger(__name__)


class PowerbiClient:
    def __init__(
        self,
        credentials: PowerbiCredentials,
    ):
        self.power_bi_client = PowerBIAPIClient(credentials)
        self.graph_api_client = MicrosoftGraphPIClient(credentials)

    def test_connection(self) -> None:
        """Use credentials & verify requesting the API doesn't raise an error"""
        self.power_bi_client._auth.refresh_token()

    def fetch_group_members(
        self,
        group_ids: Iterable[str] | None = None,
    ) -> Iterator[dict]:
        """
        Returns the list of users that are members of the given groups.
        This may contain duplicates.
        """
        if group_ids is None:
            raise ValueError("Missing group IDs to extract users")

        yield from self.graph_api_client.users_in_groups(group_ids)

    def fetch(
        self,
        asset: PowerBiAsset,
        *,
        day: date | None = None,
    ) -> Iterator[dict]:
        """
        Given a PowerBi asset, returns the corresponding data using the
        appropriate client.
        """
        if asset == PowerBiAsset.ACTIVITY_EVENTS:
            yield from self.power_bi_client.activity_events(day=day)

        elif asset == PowerBiAsset.DATAFLOWS:
            yield from self.power_bi_client.dataflows()

        elif asset == PowerBiAsset.DATASETS:
            yield from self.power_bi_client.datasets()

        elif asset == PowerBiAsset.DASHBOARDS:
            yield from self.power_bi_client.dashboards()

        elif asset == PowerBiAsset.REPORTS:
            yield from self.power_bi_client.reports()

        elif asset == PowerBiAsset.METADATA:
            yield from self.power_bi_client.metadata()

        else:
            raise ValueError(f"This asset {asset} is unknown")



from datetime import datetime

from .processor import RedshiftExtractionProcessor
from .types import AssembledQuery


def test__assemble_all() -> None:
    # build stream of query parts (intentionally out of order and mixed)
    parts = iter(
        [
            {
                "query_id": 1,
                "text": "FROM toto",
                "sequence": 3,  # third part of statement
                "sequence_count": 3,  # the third part is also the last part
            },
            {
                "query_id": 1,
                "text": "SELECT ",  # first part of statement
                "sequence": 1,
                "sequence_count": 3,
            },
            {
                "query_id": 2,
                "text": "INSERT INTO t ",
                "sequence": 1,
                "sequence_count": 2,
            },
            {
                "query_id": 1,
                "text": "* ",
                "sequence": 2,
                "sequence_count": 3,
            },  # second part of statement
            {
                "query_id": 2,
                "text": "VALUES (1,2,3);",
                "sequence": 2,
                "sequence_count": 2,
            },
        ]
    )

    processor = object.__new__(RedshiftExtractionProcessor)
    results = processor._assemble_all(parts)

    # Convert to a mapping for easy assertions
    by_query = {query.query_id: query.text for query in results}

    # Query IDs are stringified in the implementation
    expected_1 = "SELECT * FROM toto"  # "SELECT " + "* " + "FROM toto"
    assert by_query["1"] == expected_1

    # INSERT ... VALUES (...) should be normalized to DEFAULT VALUES
    assert by_query["2"].startswith("INSERT INTO t ")
    assert "DEFAULT VALUES" in by_query["2"]
    assert "VALUES (" not in by_query["2"]


def test__reconcile() -> None:
    """Test reconciliation of assembled queries with metadata"""
    # Assembled queries
    queries_text = iter(
        [
            AssembledQuery(query_id="1", text="SELECT * FROM users"),
            AssembledQuery(
                query_id="2", text="INSERT INTO logs DEFAULT VALUES"
            ),
            # for query 3 somehow it won't have corresponding metadata, but in practice this case cannot happen
            AssembledQuery(
                query_id="3", text="UPDATE products SET price = 100"
            ),
        ]
    )

    # metadata dictionaries that will serve as reference
    queries_metadata = iter(
        [
            {
                "query_id": "1",
                "aborted": 0,
                "database_id": 100,
                "database_name": "test_db",
                "end_time": datetime(2024, 1, 1, 12, 0, 0),
                "label": "query_1",
                "process_id": "proc_1",
                "start_time": datetime(2024, 1, 1, 11, 59, 0),
                "user_id": 1,
                "user_name": "test_user",
            },
            {
                "query_id": "2",
                "aborted": 0,
                "database_id": 100,
                "database_name": "test_db",
                "end_time": datetime(2024, 1, 1, 12, 1, 0),
                "label": "query_2",
                "process_id": "proc_2",
                "start_time": datetime(2024, 1, 1, 12, 0, 0),
                "user_id": 2,
                "user_name": "admin",
            },
            # Metadata for query_id "4" that has no matching query text
            {
                "query_id": "4",
                "aborted": 1,
                "database_id": 200,
                "database_name": "other_db",
                "end_time": datetime(2024, 1, 1, 12, 2, 0),
                "label": "query_4",
                "process_id": "proc_4",
                "start_time": datetime(2024, 1, 1, 12, 1, 0),
                "user_id": 3,
                "user_name": "other_user",
            },
        ]
    )

    results = list(
        RedshiftExtractionProcessor._reconcile(queries_text, queries_metadata)
    )

    # Should only return queries that have matching metadata
    assert len(results) == 2

    # query_id "1"
    result_1 = next(r for r in results if r["query_id"] == "1")
    assert result_1["query_text"] == "SELECT * FROM users"
    assert result_1["database_name"] == "test_db"
    assert result_1["user_name"] == "test_user"
    assert result_1["aborted"] == 0

    # query_id "2"
    result_2 = next(r for r in results if r["query_id"] == "2")
    assert result_2["query_text"] == "INSERT INTO logs DEFAULT VALUES"
    assert result_2["database_name"] == "test_db"
    assert result_2["user_name"] == "admin"
    assert result_2["aborted"] == 0

    # Query "3" should not appear (no matching metadata, should not happen in real life)
    assert not any(r["query_id"] == "3" for r in results)

    # Query "4" should not appear (no matching query text)
    assert not any(r["query_id"] == "4" for r in results)



import logging
import time
from collections.abc import Callable
from functools import partial
from http import HTTPStatus

from requests.exceptions import ConnectTimeout, ReadTimeout

from ....exceptions import TestConnectionException
from ....utils import (
    APIClient,
    BearerAuth,
    RequestSafeMode,
    SerializedAsset,
    fetch_all_pages,
)
from ....utils.client.api.smart_pagination import SmartPagination
from ..assets import CoalesceAsset, CoalesceQualityAsset
from .credentials import CoalesceCredentials
from .endpoint import (
    CoalesceEndpointFactory,
)
from .pagination import CoalescePagination

logger = logging.getLogger(__name__)

# increase pagination when extraction is too slow
# reduce pagination when hitting timeouts
COALESCE_PAGE_SIZE = 200
COALESCE_PAGE_SIZE_RUN_RESULTS = 1_000
COALESCE_MIN_PAGE_SIZE = 1

# raising this value might end up in remote server disconnection
COALESCE_TIMEOUT_SECONDS = 4 * 60  # 4 minutes

_MAX_ERRORS = 200

COALESCE_SAFE_MODE = RequestSafeMode(
    status_codes=(HTTPStatus.INTERNAL_SERVER_ERROR,),
    max_errors=_MAX_ERRORS,
)

RETRY_COUNT = 1
RETRY_BASE_MS = 5 * 60 * 1000  # 5 minutes
RETRY_EXCEPTIONS = (ConnectTimeout, ReadTimeout)

RETRY_EXCEPTION_MSG = """CoalesceClient raised exception '{exception}' while fetching env {environment_id},
with {{"startingFrom": {starting_from}}} and page_size of {page_size}
retrying in {retry}secs"""


def _run_result_payload(
    environment_id: str,
    result: dict,
    query_result: dict,
) -> dict:
    return {
        "environment_id": environment_id,
        "node_id": result["nodeID"],
        "node_name": result["name"],
        "test_name": query_result["name"],
        "start_time": query_result["startTime"],
        "end_time": query_result["endTime"],
        "status": query_result["status"],
        "success": query_result["success"],
        "isRunning": query_result["isRunning"],
    }


class CoalesceBearerAuth(BearerAuth):
    """Bearer Authentication for Coalesce"""

    def fetch_token(self) -> str | None:
        pass

    def __init__(self, token: str):
        self._token = token


class CoalesceClient(APIClient):
    """REST API client to extract data from Coalesce"""

    def test_connection(self) -> None:
        endpoint = CoalesceEndpointFactory.environments()
        request = self._get_paginated(endpoint=endpoint)
        try:
            next(fetch_all_pages(request, CoalescePagination))
        except Exception as e:
            raise TestConnectionException(e)

    def __init__(
        self,
        credentials: CoalesceCredentials,
    ):
        auth = CoalesceBearerAuth(token=credentials.token)
        super().__init__(
            host=credentials.host,
            auth=auth,
            safe_mode=COALESCE_SAFE_MODE,
            timeout=COALESCE_TIMEOUT_SECONDS,
        )

    def _get_paginated(
        self,
        endpoint: str,
        params: dict | None = None,
    ) -> Callable:
        return partial(
            self._get,
            retry_on_timeout=False,
            endpoint=endpoint,
            params={
                **(params or dict()),
            },
        )

    def _fetch_environments(self) -> SerializedAsset:
        endpoint = CoalesceEndpointFactory.environments()
        request = self._get_paginated(endpoint=endpoint)
        result = fetch_all_pages(request, CoalescePagination)
        return list(result)

    def _fetch_env_node_page(
        self, endpoint: str, page_size: int, starting_from: str | int | None
    ) -> tuple[dict | None, int | None]:
        response = self._call(
            method="GET",
            endpoint=endpoint,
            params={
                "detail": "true",
                "limit": page_size,
                "startingFrom": starting_from,
            },
            retry_on_timeout=False,
        ).json()
        return response.get("data"), response.get("next")

    def _fetch_env_nodes(self, environment_id: int) -> SerializedAsset:
        endpoint = CoalesceEndpointFactory.nodes(environment_id=environment_id)
        starting_from: None | str | int = None
        pagination = SmartPagination(COALESCE_PAGE_SIZE)
        result: list[dict] = []
        while True:
            try:
                data, next_page = self._fetch_env_node_page(
                    endpoint, pagination.page_size, starting_from
                )
                if data is not None:
                    result.extend(data)
                pagination.next()
                if next_page is None:
                    break
                starting_from = next_page
            except RETRY_EXCEPTIONS as e:
                if pagination.page_size <= COALESCE_MIN_PAGE_SIZE:
                    raise e
                logging.info(
                    RETRY_EXCEPTION_MSG.format(
                        exception=e,
                        starting_from=starting_from,
                        page_size=pagination.page_size,
                        environment_id=environment_id,
                        retry=RETRY_BASE_MS // 1000,
                    )
                )
                pagination.reduce_page_size()
                time.sleep(RETRY_BASE_MS // 1000)

        return [
            {
                **node,
                "environment_id": environment_id,
            }
            for node in result
        ]

    def _fetch_all_nodes(self) -> SerializedAsset:
        environments = self._fetch_environments()
        total = len(environments)
        nodes: list[dict] = []

        for index, env in enumerate(environments):
            env_id = env["id"]
            logger.info(f"Fetching nodes for env #{env_id} - {index}/{total}")
            nodes.extend(self._fetch_env_nodes(env_id))
            logger.info(f"{len(nodes)} nodes extracted so far")
        return nodes

    def _fetch_runs(self, starting_from: str | None) -> SerializedAsset:
        endpoint = CoalesceEndpointFactory.runs()
        params = {
            "orderBy": "runEndTime",
            "orderByDirection": "asc",
            "startingFrom": starting_from,
        }
        request = self._get_paginated(
            endpoint=endpoint,
            params=params,
        )
        return list(fetch_all_pages(request, CoalescePagination))

    def _fetch_run_results(self, run_id: str) -> SerializedAsset:
        endpoint = CoalesceEndpointFactory.run_results(run_id)
        result = self._get(endpoint=endpoint)
        return result["data"]

    def _run_results_by_run(
        self,
        environment_id: str,
        run_id: str,
    ) -> SerializedAsset:
        run_results: list[dict] = []
        for result in self._fetch_run_results(run_id):
            for query_result in result["queryResults"]:
                if query_result["type"] != "sqlTest":
                    continue
                run_result = _run_result_payload(
                    environment_id,
                    result,
                    query_result,
                )
                run_results.append(run_result)
        return run_results

    def _fetch_all_run_results(
        self,
        starting_from: str | None,
    ) -> SerializedAsset:
        run_results: list[dict] = []

        runs = self._fetch_runs(starting_from)
        total = len(runs)

        for index, run in enumerate(runs):
            logger.info(f"Extracting run results ({index}/{total})")
            run_id = run["id"]
            environment_id = run["environmentID"]
            current_results = self._run_results_by_run(environment_id, run_id)
            run_results.extend(current_results)
        return run_results

    def fetch(
        self,
        asset: CoalesceAsset | CoalesceQualityAsset,
        starting_from: str | None = None,
    ) -> SerializedAsset:
        """Extract the given Coalesce Asset"""
        if asset in (CoalesceAsset.NODES, CoalesceQualityAsset.NODES):
            return self._fetch_all_nodes()
        elif asset == CoalesceQualityAsset.RUN_RESULTS:
            return self._fetch_all_run_results(starting_from=starting_from)
        raise AssertionError(
            f"Asset {asset} is not supported by CoalesceClient"
        )



import logging
from collections.abc import Callable, Iterator
from http import HTTPStatus
from time import sleep
from typing import TYPE_CHECKING

from ....utils import fetch_all_pages, retry_request
from .endpoints import SigmaEndpointFactory
from .pagination import SigmaTokenPagination

if TYPE_CHECKING:
    from .client import SigmaClient

logger = logging.getLogger(__name__)

RETRY_CODES = (
    HTTPStatus.TOO_MANY_REQUESTS,
    HTTPStatus.SERVICE_UNAVAILABLE,
)

SIGMA_CONNECTION_PATH_MAX_RETRY = 1
SIGMA_CONNECTION_PATH_SLEEP_MS = 30_000

SIGMA_SOURCES_MAX_RETRY = 1
SIGMA_SOURCES_RETRY_SLEEP_MS = 30_000
SIGMA_SOURCES_REQUEST_SLEEP_S = 0.4


class SigmaSourcesTransformer:
    """Retrieves asset sources and enhances them with additional information."""

    def __init__(
        self, api_client: "SigmaClient", table_id_key: str = "inodeId"
    ):
        self.api_client = api_client
        self.table_id_key = table_id_key

    @retry_request(
        status_codes=RETRY_CODES,
        max_retries=SIGMA_CONNECTION_PATH_MAX_RETRY,
        base_ms=SIGMA_CONNECTION_PATH_SLEEP_MS,
    )
    def _get_connection_path(self, table_id: str) -> dict:
        """Retrieves the connection path for a given table id"""
        return self.api_client._get(
            endpoint=SigmaEndpointFactory.connection_path(table_id)
        )

    def _map_table_id_to_connection_path(
        self, all_sources: list
    ) -> dict[str, dict]:
        """Maps a table id to its connection and path information."""
        logger.info("Mapping table ids to connection and path information")

        unique_table_ids = {
            source[self.table_id_key]
            for asset_sources in all_sources
            for source in asset_sources.get("sources", [])
            if source["type"] == "table"
        }

        return {
            table_id: self._get_connection_path(table_id)
            for table_id in unique_table_ids
        }

    def _enhance_table_source(self, source: dict, table_to_path: dict) -> dict:
        """
        Combines a single table source with its connection and path information.
        """
        if source["type"] != "table":
            return source

        path_info = table_to_path.get(source[self.table_id_key], {})
        source["connectionId"] = path_info.get("connectionId")
        source["path"] = path_info.get("path")
        return source

    def _transform_sources(
        self, all_sources: list, table_to_path: dict
    ) -> Iterator[dict]:
        """
        Yields all sources, with table sources being enhanced with additional information.
        """
        logger.info("Merging sources with table information")

        for asset_sources in all_sources:
            enhanced_sources = [
                self._enhance_table_source(source, table_to_path)
                for source in asset_sources["sources"]
            ]

            yield {
                "asset_id": asset_sources["asset_id"],
                "sources": enhanced_sources,
            }

    @retry_request(
        status_codes=RETRY_CODES,
        max_retries=SIGMA_SOURCES_MAX_RETRY,
        base_ms=SIGMA_SOURCES_RETRY_SLEEP_MS,
    )
    def _fetch_asset_sources(
        self,
        asset_id: str,
        endpoint: Callable[[str], str],
        with_pagination: bool,
    ) -> dict:
        """Fetches sources for a single asset."""
        endpoint_url = endpoint(asset_id)

        if with_pagination:
            request = self.api_client._get_paginated(endpoint=endpoint_url)
            pages_generator = fetch_all_pages(
                request=request,
                pagination_model=SigmaTokenPagination,
                rate_limit=SIGMA_SOURCES_REQUEST_SLEEP_S,
            )
            sources = list(pages_generator)
        else:
            sources = self.api_client._get(endpoint=endpoint_url)
            sleep(SIGMA_SOURCES_REQUEST_SLEEP_S)

        return {"asset_id": asset_id, "sources": sources}

    def _get_all_sources(
        self,
        endpoint: Callable[[str], str],
        asset_ids: set[str],
        with_pagination: bool = False,
    ) -> Iterator[dict]:
        """Returns transformed sources for the given assets"""
        all_sources = [
            self._fetch_asset_sources(asset_id, endpoint, with_pagination)
            for asset_id in asset_ids
        ]

        table_to_path = self._map_table_id_to_connection_path(all_sources)

        yield from self._transform_sources(all_sources, table_to_path)

    def get_datamodel_sources(self, datamodels: list[dict]) -> Iterator[dict]:
        asset_ids = {datamodel["dataModelId"] for datamodel in datamodels}
        yield from self._get_all_sources(
            endpoint=SigmaEndpointFactory.datamodel_sources,
            asset_ids=asset_ids,
            with_pagination=True,
        )

    def get_dataset_sources(self, datasets: list[dict]) -> Iterator[dict]:
        asset_ids = {dataset["datasetId"] for dataset in datasets}
        yield from self._get_all_sources(
            endpoint=SigmaEndpointFactory.dataset_sources, asset_ids=asset_ids
        )

    def get_workbook_sources(self, workbooks: list[dict]) -> Iterator[dict]:
        asset_ids = {workbook["workbookId"] for workbook in workbooks}
        yield from self._get_all_sources(
            endpoint=SigmaEndpointFactory.workbook_sources, asset_ids=asset_ids
        )



class NoDatabaseProvidedException(Exception):
    def __init__(self):
        super().__init__("""No database eligible for extraction.
If you are using the db_allow/db_block options, please make sure to use the correct case.
        """)


class TestConnectionException(Exception):
    def __init__(self, caused_by: Exception):
        super().__init__(f"Connection could not be verified: {caused_by}")



import logging

import boto3  # type: ignore

from ...utils import SerializedAsset
from .credentials import GlueAthenaCredentials  # type: ignore

logger = logging.getLogger(__name__)

_S3_PREFIX = "s3://"
_S3_CONSOLE_BUCKETS_URL = "https://s3.console.aws.amazon.com/s3/buckets/"
_GLUE_DATABASE_NAME = "awsdatacatalog"  # https://docs.aws.amazon.com/redshift/latest/mgmt/query-editor-v2-glue.html


def _s3_to_http(url: str) -> str:
    """Convert s3:// links to the AWS console URL."""
    if not url.startswith(_S3_PREFIX):
        return url

    s3_path = url.removeprefix(_S3_PREFIX)
    parts = s3_path.split("/", 1)
    bucket = parts[0]
    object_name = parts[1] if len(parts) > 1 else ""
    return f"{_S3_CONSOLE_BUCKETS_URL}{bucket}/?prefix={object_name}"


def _format_table_description(
    location: str,
    description: str | None,
) -> str | None:
    """Append bucket location details to a table description."""

    if not location:
        return description or None

    url = _s3_to_http(location)
    location_md = f"Bucket location: [{location}]({url})"
    if not description:
        return location_md

    return f"{description}\n{location_md}"


def _table_type(table_type: str | None) -> str:
    """Normalize Glue table types into catalog-friendly labels."""
    if not table_type:
        return "TABLE"
    if table_type.upper() == "VIRTUAL_VIEW":
        return "VIEW"
    if table_type.upper() == "EXTERNAL_TABLE":
        return "EXTERNAL"
    return "TABLE"


def _parse_column(
    table_id: str,
    raw_column: dict,
    ordinal_position: int,
) -> dict | None:
    """Convert a Glue column entry into a catalog column payload."""
    column_name = raw_column.get("Name")
    if not column_name:
        return None
    return {
        "column_id": f"{table_id}.{column_name}",
        "column_name": column_name,
        # Keep Glue raw type; mapping to Catalog ColumnDataType is done at store.
        "data_type": raw_column["Type"],
        "description": raw_column.get("Comment"),
        "ordinal_position": ordinal_position,
        "table_id": table_id,
        "tags": [],
    }


def _parse_table(
    raw_table: dict,
    schema_id: str,
    table_arn: str,
) -> dict:
    table_name = raw_table["Name"]
    description = raw_table.get("Description")
    storage = raw_table.get("StorageDescriptor", {})
    location = storage["Location"]
    table_type = _table_type(raw_table.get("TableType"))
    return {
        "description": _format_table_description(location, description),
        "schema_id": schema_id,
        "table_id": table_arn,
        "table_name": table_name,
        "tags": [],
        "type": table_type,
    }


def _iter_table_columns(raw_table: dict, table_arn: str) -> SerializedAsset:
    """
    Build one ordered column list from storage columns and partition keys.

    Duplicate names are skipped so each logical column appears once.
    """
    columns: SerializedAsset = []
    ordinal_position = 0
    seen_columns: set[str] = set()
    column_entries = (
        ("column", raw_table.get("StorageDescriptor", {}).get("Columns", [])),
        ("partition key", raw_table.get("PartitionKeys", [])),
    )

    for column_kind, raw_columns in column_entries:
        for raw_column in raw_columns:
            column_name = raw_column.get("Name")
            if not column_name:
                continue
            if column_name in seen_columns:
                continue

            parsed = _parse_column(table_arn, raw_column, ordinal_position)
            if parsed is None:
                continue
            columns.append(parsed)
            seen_columns.add(column_name)
            ordinal_position += 1

    return columns


class GlueAthenaClient:
    """Fetch Glue catalog metadata for Glue/Athena."""

    def __init__(
        self,
        credentials: GlueAthenaCredentials,
    ) -> None:
        self._region = credentials.aws_region
        self._account_id = credentials.aws_account_id

        session = boto3.session.Session(
            region_name=credentials.aws_region,
            aws_access_key_id=credentials.access_key_id,
            aws_secret_access_key=credentials.access_key_secret,
        )
        self._glue_client = session.client("glue")

    def databases(self) -> SerializedAsset:
        """
        Return the single AWS Data Catalog database asset.

        AWS exposes Glue through the built-in `awsdatacatalog` database:
        https://docs.aws.amazon.com/redshift/latest/mgmt/query-editor-v2-glue.html
        """
        return [
            {
                "database_id": _GLUE_DATABASE_NAME,
                "database_name": _GLUE_DATABASE_NAME,
            }
        ]

    @property
    def _glue_arn_prefix(self) -> str:
        return f"arn:aws:glue:{self._region}:{self._account_id}"

    def _resource_arn(self, resource: str, identifier: str) -> str:
        return f"{self._glue_arn_prefix}:{resource}/{identifier}"

    def schemas(self) -> SerializedAsset:
        """
        Return schema assets by calling Glue `get_databases()`.

        Naming is intentionally normalized to schemas here, even though the
        Glue endpoint is named "databases".
        """
        schemas: SerializedAsset = []
        paginator = self._glue_client.get_paginator("get_databases")
        for page in paginator.paginate():
            # Glue's `get_databases` endpoint returns logical schemas.
            for raw_schema in page.get("DatabaseList", []):
                schema_name = raw_schema["Name"]
                schema_arn = self._resource_arn("database", schema_name)
                schemas.append(
                    {
                        "database_id": _GLUE_DATABASE_NAME,
                        "schema_id": f"{_GLUE_DATABASE_NAME}.{schema_arn}",
                        "schema_name": schema_name,
                        "description": raw_schema.get("Description"),
                        "tags": [],
                    }
                )
        return schemas

    def tables_and_columns(
        self,
        schemas: SerializedAsset,
    ) -> tuple[SerializedAsset, SerializedAsset]:
        """Collect tables and columns for each schema."""
        tables: SerializedAsset = []
        columns: SerializedAsset = []

        paginator = self._glue_client.get_paginator("get_tables")
        for schema in schemas:
            schema_name = schema["schema_name"]
            schema_id = schema["schema_id"]

            page_iterator = paginator.paginate(DatabaseName=schema_name)
            for page in page_iterator:
                for raw_table in page.get("TableList", []):
                    table_arn = self._resource_arn(
                        "table",
                        f"{schema_name}/{raw_table['Name']}",
                    )
                    table = _parse_table(raw_table, schema_id, table_arn)
                    tables.append(table)

                    table_columns = _iter_table_columns(raw_table, table_arn)
                    columns.extend(table_columns)

        return tables, columns



import pytest

from .credentials import SalesforceCredentials

TEST_KWARGS = {
    "username": "giphy",
    "password": "1312",
    "client_id": "degenie",
    "client_secret": "fautpasledire",
    "security_token": "yo",
    "base_url": "man",
}


def test_Credentials_Username_Password_token_request_payload():
    """
    When Username, Password, and Security_token are defined
    It should use Basic Username-Password Flow
    """
    creds = SalesforceCredentials(**TEST_KWARGS)

    payload = creds.token_request_payload()

    assert payload["grant_type"] == "password"
    assert payload["username"] == "giphy"
    assert payload["client_id"] == "degenie"
    assert payload["client_secret"] == "fautpasledire"
    assert payload["password"] == "1312yo"


def test_Credentials_Client_Flow_token_request_payload():
    """
    When Username, Password, and Security_token are undefined
    It should use Credential Client Flow
    """
    creds = SalesforceCredentials(
        client_id="degenie",
        client_secret="fautpasledire",
        base_url="man",
    )

    payload = creds.token_request_payload()

    assert payload["grant_type"] == "client_credentials"
    assert payload["client_id"] == "degenie"
    assert payload["client_secret"] == "fautpasledire"
    assert "username" not in payload
    assert "password" not in payload


def test_Credentials_Missing_Parameters_token_request_payload():
    """
    When using Basic Username-Password Flow all args are required.
    It should raise an error when missing one required arg
    """
    for key in (
        "username",
        "password",
        "security_token",
    ):
        kwargs = TEST_KWARGS.copy()
        del kwargs[key]
        creds = SalesforceCredentials(**kwargs)
        with pytest.raises(ValueError):
            creds.token_request_payload()



from .smart_pagination import SmartPagination


def test_smart_pagination():
    p = SmartPagination(initial_page_size=200)

    assert p.page_size == 200
    assert p._reduced_page_size is None
    assert p._counter is None
    assert p._slow_mode is False

    p.reduce_page_size()  # 200 → 20
    assert p.page_size == 20
    assert p._counter == 0
    assert p._slow_mode is True

    p.next()  # counter = 20
    assert p._counter == 20

    p.reduce_page_size()  # 20 → 2
    assert p.page_size == 2
    assert p._counter == 20  # stays 20 because we're already in slow mode
    assert p._slow_mode is True

    p.reset()

    # manual reset
    assert p._slow_mode is False
    assert p._counter is None
    assert p._reduced_page_size is None
    assert p.page_size == 200

    # running in full-speed mode
    p.next()
    p.next()
    p.next()

    assert p._slow_mode is False
    assert p._counter is None
    assert p._reduced_page_size is None
    assert p.page_size == 200

    # simulating a slow-mode once again
    p.reduce_page_size()
    assert p._slow_mode is True
    assert p._counter == 0
    assert p._reduced_page_size == 20
    assert p.page_size == 20

    for _ in range(9):
        p.next()

    # counter = 180 / 200 => still in slow mode
    assert p._slow_mode is True

    # counter reaches 200 / 200 => should get back to full speed
    p.next()
    assert p._slow_mode is False
    assert p._counter is None
    assert p._reduced_page_size is None
    assert p.page_size == 200



from ...utils import ExtractionQuery, QueryFilter, values_to_params
from ..abstract import (
    CATALOG_ASSETS,
    AbstractQueryBuilder,
    TimeFilter,
    WarehouseAsset,
)

DB_FILTERED_ASSETS = (
    *CATALOG_ASSETS,
    WarehouseAsset.COLUMN_LINEAGE,
    WarehouseAsset.QUERY,
    WarehouseAsset.VIEW_DDL,
)
QUERY_FILTERED_ASSETS = (
    WarehouseAsset.COLUMN_LINEAGE,
    WarehouseAsset.QUERY,
)


def _transient_filter(has_transient: bool | None = False) -> QueryFilter:
    """
    Build a filter controlling whether transient Snowflake objects
    should be fetched.

    The filter replaces the `{has_fetch_transient}` placeholder with either
    `TRUE` or `FALSE`, allowing queries to include or exclude transient objects
    (e.g., transient tables or schemas).
    """
    return QueryFilter(
        placeholder="has_fetch_transient",
        expression="TRUE" if has_transient else "FALSE",
    )


def _database_filter(db_list: list[str] | None, allow: bool) -> QueryFilter:
    """
    Build a filter restricting results based on allowed or blocked databases.

    This filter replaces either the {database_allowed} or {database_blocked}
    placeholder in a query. It generates an IN or NOT IN clause
    with parameterized database names.
    """
    placeholder = "database_allowed" if allow else "database_blocked"
    if not db_list:
        # replace the placeholder with dummy 'True' when list is empty
        return QueryFilter(placeholder, expression="TRUE")
    operator = "IN" if allow else "NOT IN"
    params = values_to_params(prefix=placeholder, values=db_list)
    db_params_formatted = ", ".join(f":{key}" for key in params)
    return QueryFilter(
        placeholder=placeholder,
        expression=f"database_name {operator} ({db_params_formatted})",
        params=params,
    )


def _query_blocked_filter(query_blocked: list[str] | None) -> QueryFilter:
    """
    Build a filter that excludes SQL queries matching specified blocked patterns

    This filter replaces the {query_blocked} placeholder with one
    or more "NOT ILIKE" conditions.

    The values argument is a list of case-insensitive string patterns
    to exclude from query text. SQL wildcards are supported: % and _"
    """
    placeholder = "query_blocked"
    if not query_blocked:
        # replace the placeholder with dummy 'True' when list is empty
        return QueryFilter(placeholder, expression="TRUE")
    params = values_to_params(prefix=placeholder, values=query_blocked)
    expression = " AND ".join(f"query_text NOT ILIKE :{key}" for key in params)
    return QueryFilter(
        placeholder=placeholder,
        expression=f"({expression})",
        params=params,
    )


class SnowflakeQueryBuilder(AbstractQueryBuilder):
    """
    Builds queries to extract assets from Snowflake.
    """

    def __init__(
        self,
        time_filter: TimeFilter | None = None,
        db_allowed: list[str] | None = None,
        db_blocked: list[str] | None = None,
        apply_db_filter_to_queries: bool | None = False,
        query_blocked: list[str] | None = None,
        fetch_transient: bool | None = False,
    ):
        super().__init__(time_filter=time_filter)
        self._db_allowed = _database_filter(db_allowed, allow=True)
        self._db_blocked = _database_filter(db_blocked, allow=False)
        self._query_blocked = _query_blocked_filter(query_blocked)
        self._apply_db_filter_to_queries = apply_db_filter_to_queries
        self._transient = _transient_filter(fetch_transient)

    def _get_db_filters(self, asset: WarehouseAsset) -> list[QueryFilter]:
        """
        When apply_db_filter_to_queries is False and the asset is query-related,
        returns dummy "TRUE" filters to disable database filtering.

        Otherwise, returns the configured database filters.
        """
        if (
            not self._apply_db_filter_to_queries
            and asset in QUERY_FILTERED_ASSETS
        ):
            return [
                _database_filter(None, allow=True),
                _database_filter(None, allow=False),
            ]
        return [self._db_allowed, self._db_blocked]

    def build(self, asset: WarehouseAsset) -> list[ExtractionQuery]:
        query = self.build_default(asset)

        filters: list[QueryFilter] = []
        if asset in DB_FILTERED_ASSETS:
            filters.extend(self._get_db_filters(asset))
            filters.append(self._transient)
        if asset in QUERY_FILTERED_ASSETS:
            filters.append(self._query_blocked)

        query.apply_filters(filters)
        return [query]



import logging
import re
from collections.abc import Iterator
from dataclasses import asdict
from itertools import chain

from ...utils import (
    AbstractSourceClient,
    AbstractStorage,
    SafeMode,
)
from ..abstract import (
    CATALOG_ASSETS,
    EXTERNAL_LINEAGE_ASSETS,
    QUERIES_ASSETS,
    VIEWS_ASSETS,
    AssetsMapping,
    SQLExtractionProcessor,
    WarehouseAsset,
    WarehouseAssetGroup,
)
from .query import RedshiftQueryBuilder
from .types import (
    AssembledQuery,
    LongQuery,
    QueryBuffer,
    QueryMetadata,
    QueryPart,
)

logger = logging.getLogger(__name__)

REDSHIFT_ASSETS: AssetsMapping = {
    WarehouseAssetGroup.CATALOG: CATALOG_ASSETS,
    WarehouseAssetGroup.QUERY: QUERIES_ASSETS,
    WarehouseAssetGroup.VIEW_DDL: VIEWS_ASSETS,
    WarehouseAssetGroup.ROLE: (
        WarehouseAsset.USER,
        WarehouseAsset.GROUP,
    ),
    WarehouseAssetGroup.EXTERNAL_LINEAGE: EXTERNAL_LINEAGE_ASSETS,
}


class RedshiftExtractionProcessor(SQLExtractionProcessor):
    """Extraction management for Redshift"""

    def __init__(
        self,
        client: AbstractSourceClient,
        query_builder: RedshiftQueryBuilder,
        storage: AbstractStorage,
        safe_mode: SafeMode | None = None,
    ):
        super().__init__(client, query_builder, storage, safe_mode)
        self.query_builder = query_builder

    @staticmethod
    def _normalize_insert_values(text: str) -> str:
        """Neutralize heavy VALUES payloads on INSERTs."""
        stripped = text.lstrip()
        if stripped.lower().startswith("insert into"):
            return re.sub(
                r"VALUES\s*(.*)$",
                "DEFAULT VALUES",
                text,
                flags=re.IGNORECASE | re.DOTALL,
            )
        return text

    def results(self, asset: WarehouseAsset) -> Iterator[dict]:
        """
        Return ready-to-be-exported data
        queries require some dedicated transformation & reconciliation
        """
        if asset == WarehouseAsset.QUERY:
            return self.queries()
        return self._results(asset)

    @staticmethod
    def _reconcile(
        queries_text: Iterator[AssembledQuery],
        queries_metadata: Iterator[dict],
    ) -> Iterator[dict]:
        """
        Reconcile long queries with their metadata.

        First we retrieve the queries metadata that will serve as reference
        Second we match query text with metadata & yield the combination
        """

        metadata = {}

        for query_metadata in queries_metadata:
            _metadata = QueryMetadata(**query_metadata)
            metadata[_metadata.query_id] = _metadata

        for entry in queries_text:
            if entry.query_id in metadata:
                yield {
                    "query_text": entry.text,
                    **asdict(metadata[entry.query_id]),
                }
            else:
                query_id = entry.query_id
                logger.warning(f"Query {query_id} could not be reconciled")

    @staticmethod
    def _assemble_query(entry: QueryBuffer) -> str:
        """reconstruct the full text of the query"""
        ordered_sequences = sorted(entry.parts.keys())
        return "".join(entry.parts[seq] for seq in ordered_sequences)

    def _assemble_all(
        self, query_parts: Iterator[dict]
    ) -> Iterator[AssembledQuery]:
        """
        Rebuild all full query texts from streamed parts.
        Input rows contain:
          - query_id: query identifier
          - text: fragment of the query
          - sequence: order of the fragment
          - sequence_count: total number of fragments for this query
        """
        buffers: dict[str, QueryBuffer] = dict()

        for query_part in query_parts:
            part = QueryPart(**query_part)

            entry = buffers.get(part.query_id)
            if entry is None:
                entry = QueryBuffer(expected=part.sequence_count)
                buffers[part.query_id] = entry

            entry.parts[part.sequence] = part.text

            # Emit as soon as all parts are collected for this query_id
            if len(entry.parts) >= entry.expected:
                full_text = self._assemble_query(entry)
                normalized = self._normalize_insert_values(full_text)
                yield AssembledQuery(query_id=part.query_id, text=normalized)
                del buffers[part.query_id]

        # Flush any remaining buffers (in case extraction was incomplete)
        for query_id, entry in buffers.items():
            full_text = self._assemble_query(entry)
            normalized = self._normalize_insert_values(full_text)
            yield AssembledQuery(query_id=query_id, text=normalized)

    def fetch_long_queries_text(self) -> Iterator[AssembledQuery]:
        """return the reconstructed query text (for long queries only)"""
        # To retrieve query text we cannot use LISTAGG when the query is longer than 65,535 characters.
        _query = self.query_builder.build_long_query(LongQuery.PARTS)
        raw_long_queries = self.fetch(_query)
        return self._assemble_all(raw_long_queries)

    def fetch_long_queries_metadata(self) -> Iterator[dict]:
        """return the metadata associated with long queries"""
        _query = self.query_builder.build_long_query(LongQuery.METADATA)
        return self.fetch(_query)

    def queries(self) -> Iterator[dict]:
        """redshift queries extracted (with metadata)"""
        short_queries = self._results(WarehouseAsset.QUERY)
        raw_long_queries = self.fetch_long_queries_text()
        long_queries_metadata = self.fetch_long_queries_metadata()
        long_queries = self._reconcile(raw_long_queries, long_queries_metadata)

        return chain(short_queries, long_queries)

    def extract(
        self,
        asset: WarehouseAsset,
        skip_existing: bool = False,
    ) -> str:
        """
        Process extraction for the given asset and returns the location of extracted data
        """
        asset_name = asset.value
        if skip_existing and self._storage.exists(asset_name):
            logger.info("Skipped, file already exists")
            return self._storage.path(asset_name)

        try:
            data = self.results(asset)
            return self._storage.put(asset_name, data)
        finally:
            self._client.close(dispose=True)



import logging

from ...exceptions import NoDatabaseProvidedException
from ...utils import LocalStorage, write_summary
from ...utils.constants import ENCODING_UTF8
from ..abstract import (
    CATALOG_ASSETS,
    AssetsMapping,
    WarehouseAsset,
    WarehouseAssetGroup,
    common_args,
)
from .client import GlueAthenaClient
from .credentials import GlueAthenaCredentials

logger = logging.getLogger(__name__)

GLUE_ATHENA_ASSETS: AssetsMapping = {
    WarehouseAssetGroup.CATALOG: CATALOG_ASSETS,
    WarehouseAssetGroup.ROLE: (WarehouseAsset.USER,),
}


def _credentials(params: dict) -> GlueAthenaCredentials:
    """Resolve credentials for Glue/Athena access."""
    filtered = {
        key: value for key, value in params.items() if value is not None
    }
    return GlueAthenaCredentials(**filtered)


def _write_empty_csv(storage: LocalStorage, asset: WarehouseAsset) -> str:
    """Write an empty CSV file for ingestion requirements."""
    path = storage.path(asset.value)
    with open(path, "w", encoding=ENCODING_UTF8) as file:
        file.write("")
    return path


class GlueAthenaExtractionProcessor:
    """Orchestrate Glue/Athena extraction into local storage."""

    def __init__(
        self,
        client: GlueAthenaClient,
        storage: LocalStorage,
    ) -> None:
        self._client = client
        self._storage = storage

    def extract_catalog(self) -> dict[str, str]:
        """Extract catalog assets from Glue."""
        catalog_locations: dict[str, str] = {}

        database_assets = self._client.databases()
        location = self._storage.put(
            WarehouseAsset.DATABASE.value, database_assets
        )
        catalog_locations[WarehouseAsset.DATABASE.value] = location
        logger.info(
            "Extracted %s databases to %s",
            len(database_assets),
            location,
        )

        schemas = self._client.schemas()
        if not schemas:
            raise NoDatabaseProvidedException

        location = self._storage.put(WarehouseAsset.SCHEMA.value, schemas)
        catalog_locations[WarehouseAsset.SCHEMA.value] = location
        logger.info(
            "Extracted %s schemas to %s",
            len(schemas),
            location,
        )

        tables, columns = self._client.tables_and_columns(schemas)

        location = self._storage.put(WarehouseAsset.TABLE.value, tables)
        catalog_locations[WarehouseAsset.TABLE.value] = location
        logger.info("Extracted %s tables to %s", len(tables), location)

        location = self._storage.put(WarehouseAsset.COLUMN.value, columns)
        catalog_locations[WarehouseAsset.COLUMN.value] = location
        logger.info("Extracted %s columns to %s", len(columns), location)
        return catalog_locations


def extract_all(**kwargs) -> None:
    """Run the full Glue/Athena extraction flow."""
    output_directory, _ = common_args(kwargs)

    credentials = _credentials(kwargs)
    client = GlueAthenaClient(credentials=credentials)

    storage = LocalStorage(directory=output_directory)

    extractor = GlueAthenaExtractionProcessor(
        client=client,
        storage=storage,
    )

    extractor.extract_catalog()
    # Glue/Athena does not provide warehouse users/roles in this extractor.
    # We still emit an empty USER file to keep role ingestion contracts
    # consistent with other technologies where this file is mandatory.
    _write_empty_csv(storage, WarehouseAsset.USER)

    write_summary(
        output_directory,
        storage.stored_at_ts,
        client_name="glue_athena",
    )



from argparse import ArgumentParser

from castor_extractor.utils import parse_filled_arguments  # type: ignore
from castor_extractor.visualization import looker_studio  # type: ignore


def main():
    parser = ArgumentParser(
        description="Extract data from Looker Studio and related BigQuery source queries"
    )
    parser.add_argument("-o", "--output", help="Directory to write to")

    mode_group = parser.add_argument_group("Execution mode")
    mode_group.add_argument(
        "--source-queries-only",
        action="store_true",
        default=False,
        help="If selected, only extracts BigQuery source queries (bypasses Looker Studio extraction)",
    )
    mode_group.add_argument(
        "--skip-view-activity-logs",
        action="store_true",
        default=False,
        help="Skip extraction of activity logs (use if credentials lack required scopes)",
    )

    looker_group = parser.add_argument_group("Looker Studio extraction")
    looker_group.add_argument(
        "-c",
        "--credentials",
        help="File path to Service Account credentials with Looker Studio access",
    )
    looker_group.add_argument(
        "-a",
        "--admin-email",
        help="Email of a Google Workspace user with admin access",
    )
    looker_group.add_argument(
        "--users-file-path",
        help=(
            "Optional path to a JSON file with user email addresses "
            'as a list of strings (e.g. ["foo@bar.com", "fee@bar.com"]). '
            "If provided, only extracts assets owned by the specified users."
        ),
    )

    bigquery_group = parser.add_argument_group("Source queries extraction")
    bigquery_group.add_argument(
        "-b",
        "--bigquery-credentials",
        help=(
            "File path to Service Account credentials with BigQuery access. "
            "Can point to the same file as Looker Studio credentials."
        ),
    )
    bigquery_group.add_argument(
        "--db-allowed",
        nargs="*",
        help="Optional list of GCP projects to allow for source queries extraction",
    )
    bigquery_group.add_argument(
        "--db-blocked",
        nargs="*",
        help="Optional list of GCP projects to block from source queries extraction",
    )

    looker_studio.extract_all(**parse_filled_arguments(parser))



from pydantic import ConfigDict, Field
from pydantic.alias_generators import to_camel

from ....utils import (
    FetchNextPageBy,
    PaginationModel,
)


class PowerBIAPIPagination(PaginationModel):
    """Handles the pagination of Power BI REST API calls for activity events"""

    model_config = ConfigDict(
        alias_generator=to_camel,
        populate_by_name=True,
        from_attributes=True,
    )

    fetch_by: FetchNextPageBy = FetchNextPageBy.URL

    activity_event_entities: list
    continuation_uri: str | None = None
    last_result_set: bool = False

    def is_last(self) -> bool:
        return self.last_result_set

    def next_page_payload(self) -> str | None:
        return self.continuation_uri

    def page_results(self) -> list:
        return self.activity_event_entities


class GraphAPIPagination(PaginationModel):
    """
    Handles the pagination of Graph API calls.
    The response is expected to be of the form:
    ```
    {
        "@odata.context":"<the URL that was called>",
        "@odata.nextLink":"<the URL to the next page>",
        "value":[
            {<info user 1>},
            {<info user 2>},
            {<info user 3>},
       ]
    }
    ```

    The "@odata.nextLink" key is optional.
    """

    fetch_by: FetchNextPageBy = FetchNextPageBy.URL

    value: list = Field(default_factory=list)
    next_url: str | None = Field(
        validation_alias="@odata.nextLink",
        default=None,
    )

    def is_last(self) -> bool:
        return self.next_url is None

    def next_page_payload(self) -> str | None:
        return self.next_url

    def page_results(self) -> list:
        return self.value



from unittest.mock import patch

import pytest

from . import PowerbiClient, PowerbiCredentials
from .authentication import msal


@pytest.fixture
def mock_msal():
    with patch.object(msal, "ConfidentialClientApplication") as mock_app:
        mock_app.return_value.acquire_token_for_client.return_value = {
            "access_token": "fake_token"
        }
        yield mock_app


@pytest.fixture
def client(mock_msal):
    creds = PowerbiCredentials(tenant_id="", client_id="", secret="pwd")
    return PowerbiClient(creds)


def test_test_connection(client):
    with patch.object(
        client.power_bi_client._auth, "refresh_token"
    ) as mock_refresh:
        client.test_connection()
        mock_refresh.assert_called_once()



from ...utils import ExtractionQuery
from .client import MSSQLClient


class FakeConnection:
    closed = False

    def execute(self, *_args, **_kwargs):
        return [{"ok": 1}]

    def close(self):
        self.closed = True


class FakeEngine:
    def __init__(self):
        self.connection = FakeConnection()

    def connect(self):
        return self.connection

    def dispose(self):
        pass


def test_build_uri_and_database_resolution(monkeypatch):
    """
    Single test covering:
    - default_db present / absent
    - query.database present / absent
    - ensuring the correct DB is used in the final engine URI
    """

    created_uris: list[str] = []

    def fake_create_engine(uri, **_kwargs):
        created_uris.append(uri)
        return FakeEngine()

    # Patch create_engine where it is used (inside client.py)
    monkeypatch.setattr(
        "source.packages.extractor.castor_extractor.warehouse.sqlserver.client.create_engine",
        fake_create_engine,
    )

    base_credentials = {
        "host": "sql.example.local",
        "port": 1433,
        "user": "user",
        "password": "password",
    }

    # Case 1: no default_db, no query.database → base URI
    client = MSSQLClient(base_credentials)
    list(client.execute(ExtractionQuery("SELECT 1", params={}, database=None)))
    assert created_uris[-1] == client._uri

    # Case 2: default_db provided, no query.database → default_db used
    client = MSSQLClient({**base_credentials, "default_db": "default_db"})
    list(client.execute(ExtractionQuery("SELECT 1", params={}, database=None)))
    assert created_uris[-1].endswith("/default_db")

    # Case 3: default_db provided, query.database provided → query.database wins
    client = MSSQLClient({**base_credentials, "default_db": "default_db"})
    list(
        client.execute(
            ExtractionQuery("SELECT 1", params={}, database="explicit_db")
        )
    )
    assert created_uris[-1].endswith("/explicit_db")

    # Case 4: no default_db, query.database provided → query.database used
    client = MSSQLClient(base_credentials)
    list(
        client.execute(
            ExtractionQuery("SELECT 1", params={}, database="explicit_db")
        )
    )
    assert created_uris[-1].endswith("/explicit_db")


def test_get_databases(monkeypatch):
    client = MSSQLClient(
        {"host": "h", "user": "u", "password": "p"},
        db_allowed=["db1", "db2"],
    )

    def fake_execute(_query):
        # Simulate sys.databases output
        yield from [
            {"name": "master"},
            {"name": "db1"},
            {"name": "db2"},
            {"name": "db3"},
        ]

    monkeypatch.setattr(client, "execute", fake_execute)

    # Make only db1 eligible for query extraction
    monkeypatch.setattr(
        client,
        "_has_queries_permissions",
        lambda database: database == "db1",
    )

    # Default behavior: system DBs removed, allow filter applied
    assert client.get_databases() == ["db1", "db2"]

    # Query extraction behavior: further restricted by query permissions
    assert client.get_databases(with_query_store_enabled=True) == ["db1"]



from abc import abstractmethod
from collections.abc import Callable, Iterator, Sequence
from enum import Enum
from itertools import chain
from typing import (
    Generic,
    TypeVar,
)

DEFAULT_PER_PAGE = 100

T = TypeVar("T")


class PagerLogger:
    def on_page(self, page: int, count: int):
        pass

    def on_success(self, page: int, total: int):
        pass


class PagerStopStrategy(Enum):
    """Strategy for stopping the pager"""

    EMPTY_PAGE = "EMPTY_PAGE"
    LESS_RESULTS_THAN_ASKED = "LESS_RESULTS_THAN_ASKED"


class AbstractPager(Generic[T]):
    def all(self, per_page: int = DEFAULT_PER_PAGE) -> list[T]:
        """Returns all data provided by the callback as a list"""
        return list(chain.from_iterable(self.iterator(per_page=per_page)))

    @abstractmethod
    def iterator(self, per_page: int) -> Iterator[Sequence[T]]:
        pass

    @staticmethod
    def should_stop(nb_results: int, per_page: int, stop_on_empty_page: bool):
        is_empty = nb_results == 0
        is_partial_page = nb_results < per_page
        should_stop = is_empty if stop_on_empty_page else is_partial_page
        return should_stop


class Pager(AbstractPager[T]):
    def __init__(
        self,
        callback: Callable[[int, int], Sequence[T]],
        *,
        logger: PagerLogger | None = None,
        start_page: int = 1,
        stop_strategy: PagerStopStrategy = PagerStopStrategy.EMPTY_PAGE,
    ):
        self._callback = callback
        self._logger = logger or PagerLogger()
        self._start_page = start_page
        self._stop_strategy = stop_strategy

    def iterator(
        self,
        per_page: int = DEFAULT_PER_PAGE,
    ) -> Iterator[Sequence[T]]:
        """Yields data provided by the callback as a list page by page"""
        page = self._start_page
        total_results = 0

        stop_on_empty_page = self._stop_strategy == PagerStopStrategy.EMPTY_PAGE

        while True:
            results = self._callback(page, per_page)
            nb_results = len(results)
            total_results += nb_results

            self._logger.on_page(page, nb_results)
            if results:
                yield results

            stop = self.should_stop(nb_results, per_page, stop_on_empty_page)

            if stop:
                break

            page += 1

        self._logger.on_success(page, total_results)



import logging
from argparse import ArgumentParser

from castor_extractor.warehouse import glue_athena  # type: ignore

logging.basicConfig(level=logging.INFO, format="%(levelname)s - %(message)s")


def main():
    parser = ArgumentParser()

    parser.add_argument("--access-key-id", help="AWS access key id")
    parser.add_argument("--access-key-secret", help="AWS access key secret")
    parser.add_argument("--aws-region", help="AWS region")
    parser.add_argument("--aws-account-id", help="AWS account id")

    parser.add_argument("-o", "--output", help="Directory to write to")

    args = parser.parse_args()

    glue_athena.extract_all(
        access_key_id=args.access_key_id,
        access_key_secret=args.access_key_secret,
        aws_region=args.aws_region,
        aws_account_id=args.aws_account_id,
        output_directory=args.output,
    )



from .asset import (
    EXTERNAL_LINEAGE_ASSETS,
    AssetsMapping,
    WarehouseAsset,
    WarehouseAssetGroup,
    extractable_assets,
)

COOL_TECHNOLOGY_ASSETS: AssetsMapping = {
    WarehouseAssetGroup.CATALOG: (
        WarehouseAsset.DATABASE,
        WarehouseAsset.SCHEMA,
    ),
    WarehouseAssetGroup.ROLE: (WarehouseAsset.USER,),
    WarehouseAssetGroup.EXTERNAL_LINEAGE: EXTERNAL_LINEAGE_ASSETS,
}


def test_extractable_asset_groups():
    # Default behavior: all extractable groups, external lineage excluded
    all_assets = set(extractable_assets(COOL_TECHNOLOGY_ASSETS))
    assert all_assets == {
        WarehouseAsset.DATABASE,
        WarehouseAsset.SCHEMA,
        WarehouseAsset.USER,
    }

    # With skip_groups: ROLE explicitly skipped, external lineage still excluded
    all_assets = set(
        extractable_assets(
            COOL_TECHNOLOGY_ASSETS,
            skip_groups={WarehouseAssetGroup.ROLE},
        )
    )
    assert all_assets == {
        WarehouseAsset.DATABASE,
        WarehouseAsset.SCHEMA,
    }



import logging
from abc import abstractmethod
from collections.abc import Callable, Iterator
from enum import Enum
from functools import partial
from time import sleep

from pydantic import BaseModel

logger = logging.getLogger(__name__)


class FetchNextPageBy(Enum):
    """
    Enum to pick which APIClient._call() argument we want
    to use for calling the next page in the pagination.
    Supported arguments are :
    - params (PAYLOAD)
    - endpoint (URL)
    """

    PAYLOAD = "pagination_params"
    URL = "endpoint"


class PaginationModel(BaseModel):
    """
    Base abstract class defining a pagination model

    By implementing the 3 abstract methods below, enables
    to fetch all elements of a Paginated API by using the
    `fetch_all_pages` method
    """

    fetch_by: FetchNextPageBy = FetchNextPageBy.PAYLOAD
    current_page_payload: dict | None = None

    @abstractmethod
    def is_last(self) -> bool:
        """Stopping condition for the pagination"""
        pass

    @abstractmethod
    def next_page_payload(self) -> dict | str | None:
        """Payload enabling to generate the request for the next page"""
        pass

    @abstractmethod
    def page_results(self) -> list:
        """List of results of the current page"""
        pass

    def next_page_parameters(self) -> dict:
        return {self.fetch_by.value: self.next_page_payload()}


def fetch_all_pages(
    request: Callable,
    pagination_model: type[PaginationModel],
    rate_limit: float | None = None,
) -> Iterator:
    """
    Method to return all results of a Paginated API based on the
    pagination model and the first request call
    """
    page_number = 1
    response_payload = request()

    paginated_response = pagination_model(**response_payload)

    while not paginated_response.is_last():
        logger.debug(f"Fetching page number {page_number}")
        yield from paginated_response.page_results()
        next_page_parameters = paginated_response.next_page_parameters()
        request_with_pagination = partial(request, **next_page_parameters)
        if rate_limit:
            sleep(rate_limit)
        paginated_response = pagination_model(
            current_page_payload=next_page_parameters,
            **request_with_pagination(),
        )
        page_number += 1

    # send last page's results
    logger.debug(f"Fetching page number {page_number}")
    yield from paginated_response.page_results()



import pytest

from .client import (
    GlueAthenaClient,
    _format_table_description,
    _iter_table_columns,
    _parse_column,
    _parse_table,
    _s3_to_http,
    _table_type,
)


@pytest.mark.parametrize(
    ("url", "expected"),
    [
        ("https://example.com/path", "https://example.com/path"),
        (
            "s3://my-bucket/path/to/object.csv",
            "https://s3.console.aws.amazon.com/s3/buckets/"
            "my-bucket/?prefix=path/to/object.csv",
        ),
        (
            "s3://my-bucket",
            "https://s3.console.aws.amazon.com/s3/buckets/my-bucket/?prefix=",
        ),
        (
            "s3://my-bucket/folder/",
            "https://s3.console.aws.amazon.com/s3/buckets/"
            "my-bucket/?prefix=folder/",
        ),
    ],
)
def test__s3_to_http(url: str, expected: str):
    assert _s3_to_http(url) == expected


def test__format_table_description_without_description():
    location = "s3://bucket/folder/"
    expected = (
        "Bucket location: [s3://bucket/folder/]"
        "(https://s3.console.aws.amazon.com/s3/buckets/"
        "bucket/?prefix=folder/)"
    )
    assert _format_table_description(location, None) == expected


def test__format_table_description_with_description():
    location = "s3://bucket/folder/"
    description = "Table description"
    expected = (
        "Table description\n"
        "Bucket location: [s3://bucket/folder/]"
        "(https://s3.console.aws.amazon.com/s3/buckets/"
        "bucket/?prefix=folder/)"
    )
    assert _format_table_description(location, description) == expected


def test__format_table_description_empty_location_without_description():
    assert _format_table_description("", None) is None


def test__format_table_description_empty_location_with_description():
    assert (
        _format_table_description("", "Table description")
        == "Table description"
    )


def test__table_type():
    assert _table_type(None) == "TABLE"
    assert _table_type("EXTERNAL_TABLE") == "EXTERNAL"
    assert _table_type("VIRTUAL_VIEW") == "VIEW"
    assert _table_type("random") == "TABLE"


def test__parse_column_missing_name():
    assert _parse_column("table_arn", {}, 0) is None


def test__parse_column_with_name():
    raw = {"Name": "col_a", "Type": "varchar", "Comment": "desc"}
    parsed = _parse_column("table_arn", raw, 2)
    assert parsed == {
        "column_id": "table_arn.col_a",
        "column_name": "col_a",
        "data_type": "varchar",
        "description": "desc",
        "ordinal_position": 2,
        "table_id": "table_arn",
        "tags": [],
    }


def test__resource_arn():
    client = GlueAthenaClient.__new__(GlueAthenaClient)
    client._region = "eu-west-1"
    client._account_id = "123"

    assert client._resource_arn("database", "db1") == (
        "arn:aws:glue:eu-west-1:123:database/db1"
    )
    assert client._resource_arn("table", "db1/table1") == (
        "arn:aws:glue:eu-west-1:123:table/db1/table1"
    )


def test_databases_is_hard_coded_catalog():
    client = GlueAthenaClient.__new__(GlueAthenaClient)
    assert client.databases() == [
        {"database_id": "awsdatacatalog", "database_name": "awsdatacatalog"}
    ]


def test__parse_table():
    raw_table = {
        "Name": "orders",
        "Description": "Orders table",
        "TableType": "EXTERNAL_TABLE",
        "StorageDescriptor": {"Location": "s3://bucket/orders/"},
    }

    parsed = _parse_table(raw_table, "schema-id", "table-arn")
    assert parsed == {
        "table_id": "table-arn",
        "schema_id": "schema-id",
        "table_name": "orders",
        "description": (
            "Orders table\n"
            "Bucket location: [s3://bucket/orders/]"
            "(https://s3.console.aws.amazon.com/s3/buckets/"
            "bucket/?prefix=orders/)"
        ),
        "tags": [],
        "type": "EXTERNAL",
    }


def test__iter_table_columns_keeps_order_and_skips_duplicates():
    raw_table = {
        "StorageDescriptor": {
            "Columns": [
                {"Name": "id", "Type": "int"},
                {"Name": "ds", "Type": "string"},
            ]
        },
        "PartitionKeys": [
            {"Name": "ds", "Type": "string"},
            {"Name": "country", "Type": "varchar"},
        ],
    }

    parsed = _iter_table_columns(raw_table, "table-arn")
    assert [column["column_name"] for column in parsed] == [
        "id",
        "ds",
        "country",
    ]
    assert [column["ordinal_position"] for column in parsed] == [0, 1, 2]



from .client import MicrosoftGraphPIClient, PowerbiClient
from .credentials import (
    CLIENT_APP_BASE,
    DEFAULT_SCOPES,
    GRAPH_API_BASE_PATH,
    REST_API_BASE_PATH,
    PowerbiCertificate,
    PowerbiCredentials,
)



import logging
from collections.abc import Iterable

from ...utils import (
    current_timestamp,
    deep_serialize,
    get_output_filename,
    write_json,
    write_summary,
)
from .assets import LookerStudioAsset
from .client import LookerStudioClient
from .parameters import ExtractionParameters, set_extraction_parameters

logger = logging.getLogger(__name__)


def iterate_all_data(
    client: LookerStudioClient, assets: Iterable[LookerStudioAsset]
) -> Iterable[tuple[LookerStudioAsset, list | dict]]:
    """Extracts Looker Studio data depending on the parameters."""
    for asset in assets:
        logger.info(f"Extracting {asset.name} from API")
        data = list(client.fetch(asset))
        yield asset, deep_serialize(data)
        logger.info(f"Extracted {len(data)} {asset.name} from API")


def _extractable_assets(
    parameters: ExtractionParameters,
) -> list[LookerStudioAsset]:
    """Returns the list of assets to extract based on the parameters."""
    if parameters.has_source_queries_only:
        return [LookerStudioAsset.SOURCE_QUERIES]

    assets = [
        LookerStudioAsset.ASSETS,
        LookerStudioAsset.SOURCE_QUERIES,
    ]
    if parameters.has_view_activity_logs:
        assets.append(LookerStudioAsset.VIEW_ACTIVITY)
    return assets


def extract_all(**kwargs) -> None:
    """
    Extracts data from Looker Studio and stores the output files locally under
    the given output_directory.
    """
    parameters = set_extraction_parameters(kwargs)

    client = LookerStudioClient(
        credentials=parameters.looker_studio_credentials,
        bigquery_credentials=parameters.bigquery_credentials,
        user_emails=parameters.user_emails,
        database_allowed=parameters.db_allowed,
        database_blocked=parameters.db_blocked,
    )
    output_dir = parameters.output_directory
    ts = current_timestamp()

    assets = _extractable_assets(parameters)
    for key, data in iterate_all_data(client, assets):
        filename = get_output_filename(
            name=key.name.lower(), output_directory=output_dir, ts=ts
        )
        write_json(filename, data)

    write_summary(output_dir, ts)



import json
import logging
from typing import Any, cast

from pydantic import BaseModel, field_validator, model_validator

from ...utils import (
    OUTPUT_DIR,
    from_env,
)
from .client import LookerStudioCredentials
from .constants import APPLICATION_CREDENTIALS, LOOKER_STUDIO_ADMIN_EMAIL

logger = logging.getLogger(__name__)

_BIGQUERY_CREDENTIALS_REQUIRED_ERROR = (
    "BigQuery credentials are required when using --source-queries-only mode"
)
_EMPTY_USERS_FILE_ERROR = "The users file must contain at least one user email"
_LOOKER_STUDIO_CREDENTIALS_REQUIRED_ERROR = "Looker Studio credentials are required unless using the --source-queries-only mode"
_USERS_FILE_NOT_A_LIST_ERROR = "The users file must be a list"
_USER_EMAIL_NOT_STRING_ERROR = "All items in the users list must be strings"


class ExtractionParameters(BaseModel):
    bigquery_credentials: dict | None = None
    db_allowed: set[str] | None = None
    db_blocked: set[str] | None = None
    has_source_queries_only: bool
    has_view_activity_logs: bool
    looker_studio_credentials: LookerStudioCredentials | None = None
    output_directory: str
    user_emails: list[str] | None = None

    @field_validator("db_allowed", "db_blocked", mode="before")
    @classmethod
    def _transform_db_list_to_set(
        cls, db_list: list[str] | None
    ) -> set[str] | None:
        return None if db_list is None else set(db_list)

    @field_validator("user_emails", mode="before")
    @classmethod
    def _validate_user_emails(
        cls, user_emails: list[str] | None
    ) -> list[str] | None:
        """
        Raises an error if the user emails are not in the expected format
        (list of strings), or if the list is empty.
        """
        if user_emails is None:
            return user_emails

        if not isinstance(user_emails, list):
            raise TypeError(_USERS_FILE_NOT_A_LIST_ERROR)

        if len(user_emails) == 0:
            raise ValueError(_EMPTY_USERS_FILE_ERROR)

        if not all(isinstance(email, str) for email in user_emails):
            raise TypeError(_USER_EMAIL_NOT_STRING_ERROR)

        return user_emails

    @model_validator(mode="after")
    def validate_credentials_requirements(self):
        """
        Validates that the right credentials are provided based on the extraction mode.
        """
        if self.has_source_queries_only:
            # Source queries only mode - requires BigQuery credentials
            if self.bigquery_credentials is None:
                raise ValueError(_BIGQUERY_CREDENTIALS_REQUIRED_ERROR)
            return self

        if self.looker_studio_credentials is None:
            raise ValueError(_LOOKER_STUDIO_CREDENTIALS_REQUIRED_ERROR)

        return self


def _load_optional_file(file_path: str | None) -> Any | None:
    """Attempts to load a JSON file from the given path."""
    if not file_path:
        return None

    logger.info(f"Loading file {file_path}")
    with open(file_path) as file:
        return json.load(file)


def _load_credentials_or_none(file_path: str | None) -> dict | None:
    """
    Attempts to load the Service Account credentials from a file or from the
    special GOOGLE_APPLICATION_CREDENTIALS variable, if it was provided.
    """
    path = file_path or from_env(APPLICATION_CREDENTIALS, allow_missing=True)
    credentials = _load_optional_file(path)
    return cast(dict, credentials)


def _credentials_or_none(params: dict) -> LookerStudioCredentials | None:
    """
    Builds the Looker Studio credentials by combining the Service Account
    credentials with the admin email. Returns None if credentials are not available.
    """
    path = params.get("credentials")
    credentials = _load_credentials_or_none(path)
    if not credentials:
        return None

    admin_email = params.get("admin_email") or from_env(
        LOOKER_STUDIO_ADMIN_EMAIL
    )
    credentials["admin_email"] = admin_email
    has_view_activity_logs = not params["skip_view_activity_logs"]
    credentials["has_view_activity_logs"] = has_view_activity_logs
    return LookerStudioCredentials(**credentials)


def _bigquery_credentials_or_none(params: dict) -> dict | None:
    """Extracts optional GCP credentials to access BigQuery"""
    path = params.get("bigquery_credentials")
    return _load_credentials_or_none(path)


def set_extraction_parameters(parsed_args: dict) -> ExtractionParameters:
    """
    Builds the ExtractionParameters from the parsed arguments and/or environment variables.
    """
    credentials = _credentials_or_none(parsed_args)
    bigquery_credentials = _bigquery_credentials_or_none(parsed_args)

    db_allowed = parsed_args.get("db_allowed")
    db_blocked = parsed_args.get("db_blocked")
    output_directory = parsed_args.get("output") or from_env(OUTPUT_DIR)
    has_view_activity_logs = bool(not parsed_args["skip_view_activity_logs"])
    source_queries_only = parsed_args.get("source_queries_only", False)

    users_file_path = parsed_args.get("users_file_path")
    user_emails = _load_optional_file(users_file_path)

    return ExtractionParameters(
        bigquery_credentials=bigquery_credentials,
        db_allowed=db_allowed,
        db_blocked=db_blocked,
        has_view_activity_logs=has_view_activity_logs,
        looker_studio_credentials=credentials,
        output_directory=output_directory,
        has_source_queries_only=source_queries_only,
        user_emails=user_emails,
    )



import logging
from collections.abc import Iterator
from http import HTTPStatus
from typing import Any, cast

import requests
from requests import HTTPError

from .....utils import (
    JsonType,
    RequestSafeMode,
    SerializedAsset,
    handle_response,
)
from ...assets import EXPORTED_FIELDS, MetabaseAsset
from ...errors import MetabaseLoginError, SuperuserCredentialsRequired
from ...types import IdsType
from ..shared import DETAILS_KEY, get_dbname_from_details
from .credentials import MetabaseApiCredentials

logger = logging.getLogger(__name__)

# Safe mode
VOLUME_IGNORED = 5
IGNORED_ERROR_CODES = (
    HTTPStatus.BAD_REQUEST,
    HTTPStatus.NOT_FOUND,
)
METABASE_SAFE_MODE = RequestSafeMode(
    max_errors=VOLUME_IGNORED,
    status_codes=IGNORED_ERROR_CODES,
)

URL_TEMPLATE = "{base_url}/api/{endpoint}"

ROOT_KEY = "root"
CARDS_KEY = "dashcards"
CARDS_KEY_DEPRECATED = "ordered_cards"  # prior to v0.48 of Metabase
DATA_KEY = "data"


class ApiClient:
    """
    Connect to Metabase API and fetch main assets.
    Superuser credentials are required.
    https://www.metabase.com/docs/latest/api-documentation.html
    """

    def __init__(
        self,
        credentials: MetabaseApiCredentials,
        safe_mode: RequestSafeMode | None = None,
    ):
        self.base_url = credentials.base_url

        self._credentials = credentials
        self._session = requests.Session()
        self._session_id = self._login()
        self.safe_mode = safe_mode or METABASE_SAFE_MODE
        self._check_permissions()  # verify that the given user is superuser

    @staticmethod
    def name() -> str:
        """return the name of the client"""
        return "Metabase/API"

    def _url(self, endpoint: str) -> str:
        return URL_TEMPLATE.format(
            base_url=self.base_url,
            endpoint=endpoint,
        )

    def _headers(self) -> dict:
        return {
            "X-Metabase-Session": self._session_id,
            "Content-type": "application/x-www-form-urlencoded",
        }

    @staticmethod
    def _answer(response: Any):
        answer = response
        if isinstance(answer, dict) and DATA_KEY in answer:
            # v0.41 of Metabase introduced embedded data for certain calls
            # {'data': [{ }, ...] , 'total': 15, 'limit': None, 'offset': None}"
            return answer[DATA_KEY]
        return answer

    def _call(self, endpoint: str) -> JsonType:
        url = self._url(endpoint)
        headers = self._headers()
        response = self._session.get(url=url, headers=headers)
        response = handle_response(response, safe_mode=self.safe_mode)
        return self._answer(response)

    def _check_permissions(self) -> None:
        try:
            # This endpoint requires superuser credentials
            self._call("collection/graph")
        except HTTPError as err:
            if err.response.status_code == 403:  # forbidden
                raise SuperuserCredentialsRequired(
                    credentials_info=self._credentials,
                    error_details=err.args,
                )
            raise

    def _login(self) -> str:
        url = self._url("session")
        payload = {
            "username": self._credentials.user,
            "password": self._credentials.password,
        }
        response = self._session.post(url, json=payload)
        logger.info(f"Getting session_id: {response.json()}")

        if not response.json().get("id"):
            raise MetabaseLoginError(
                credentials_info=self._credentials,
                error_details=response.json(),
            )

        return response.json()["id"]

    def _fetch_ids(self, asset: MetabaseAsset) -> IdsType:
        ids: IdsType = []
        results = self._call(endpoint=asset.name.lower())
        for res in cast(list, results):
            assert isinstance(res, dict)
            ids.append(res["id"])
        return ids

    def _dashboards(self) -> Iterator[dict]:
        collection_ids = self._fetch_ids(MetabaseAsset.COLLECTION)
        for _id in collection_ids:
            collection = self._call(f"collection/{_id}/items?models=dashboard")
            if not collection:
                continue

            seen_dashboard_ids: set[int] = set()

            for dashboard in cast(SerializedAsset, collection):
                if dashboard.get("model") != "dashboard":
                    # This is to maintain compatibility with older versions
                    # where ?models=dashboard has no effects
                    continue

                dashboard_id = dashboard.get("id")
                if not dashboard_id:
                    continue

                if dashboard_id not in seen_dashboard_ids:
                    seen_dashboard_ids.add(dashboard_id)
                    yield cast(dict, self._call(f"dashboard/{dashboard_id}"))

    @staticmethod
    def _collection_specifics(collections: SerializedAsset) -> SerializedAsset:
        # remove the root folder
        def _is_not_root(collection: dict) -> bool:
            return collection.get("id") != ROOT_KEY

        return list(filter(_is_not_root, collections))

    @staticmethod
    def _database_specifics(databases: SerializedAsset) -> SerializedAsset:
        for db in databases:
            # superuser privileges are mandatory, this field should be present
            assert DETAILS_KEY in db
            details = db[DETAILS_KEY]
            db["dbname"] = get_dbname_from_details(details)

        return databases

    @staticmethod
    def _dashboard_cards(dashboards: SerializedAsset) -> Iterator[dict]:
        for d in dashboards:
            d_cards = d.get(CARDS_KEY) or d.get(CARDS_KEY_DEPRECATED) or []
            yield from d_cards

    def fetch(self, asset: MetabaseAsset) -> SerializedAsset:
        """fetches the given asset"""
        if asset == MetabaseAsset.DASHBOARD:
            assets = list(self._dashboards())

        elif asset == MetabaseAsset.DASHBOARD_CARDS:
            dashboards = list(self._dashboards())
            assets = list(self._dashboard_cards(dashboards))

        else:
            answer = self._call(asset.name.lower())
            assets = cast(list, answer)

        if asset == MetabaseAsset.DATABASE:
            assets = self._database_specifics(assets)

        if asset == MetabaseAsset.COLLECTION:
            assets = self._collection_specifics(assets)

        logger.info(f"Fetching {asset.name} ({len(assets)} results)")

        # keep interesting fields
        return [
            {key: e.get(key) for key in EXPORTED_FIELDS[asset]} for e in assets
        ]
